#include "GLWindow.hpp"

#include "GLHelper.hpp"
#include <iostream>
#include <stdexcept>

GLWindow::GLWindow(int width, int height, const std::string& title)
  : m_window(nullptr),
    m_texture(0)
{
  // Init GLFW
  if (!glfwInit())
    throw std::runtime_error("GLFW initialization failed");

  // Create window
  glfwWindowHint(GLFW_RESIZABLE, GL_FALSE);
  m_window = glfwCreateWindow(width, height, title.c_str(), nullptr, nullptr);
  if(!m_window)
  {
    glfwTerminate();
    throw std::runtime_error("GLFW window creation failed");
  }
  glfwMakeContextCurrent(m_window);
  glfwSwapInterval(0);

  // Init GLEW
  GLenum err = glewInit();
  if(err != GLEW_OK)
  {
    glfwTerminate();
    throw std::runtime_error("GLEW initialization failed");
  }

  // Create resources
  glGenTextures(1, &m_texture);
  glBindTexture(GL_TEXTURE_2D, m_texture);

  // Enable texture mapping
  glEnable(GL_TEXTURE_2D);
  glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_NEAREST);
  glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_NEAREST);
  glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP);
  glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP);
}

GLWindow::~GLWindow()
{
  glDeleteTextures(1, &m_texture);
  glfwTerminate();
}

bool GLWindow::Run()
{
  glfwPollEvents();

  // Check for openGL errors
  GLenum err = glGetError();
  if (err != GL_NO_ERROR)
    std::cout << "OpenGL error: " << GLGetErrorString(err) << std::endl;

  return !glfwWindowShouldClose(m_window);
}

void GLWindow::Clear()
{
  glClear(GL_COLOR_BUFFER_BIT);
}

void GLWindow::Present()
{
  glfwSwapBuffers(m_window);
}

void GLWindow::DrawTexture(const uint8_t* data, int width, int height)
{
  // Set texture data
  glPixelStorei(GL_UNPACK_ALIGNMENT, 1);
  glTexImage2D(GL_TEXTURE_2D, 0, GL_LUMINANCE8, width, height, 0,
               GL_LUMINANCE, GL_UNSIGNED_BYTE, data);
  glActiveTexture(GL_TEXTURE0);

  // Draw
  glBegin(GL_QUADS);
    glTexCoord2f(0.f, 1.f); glVertex2f(-1.f, -1.f);
    glTexCoord2f(1.f, 1.f); glVertex2f(1.f, -1.f);
    glTexCoord2f(1.f, 0.f); glVertex2f(1.f, 1.f);
    glTexCoord2f(0.f, 0.f); glVertex2f(-1.f, 1.f);
  glEnd();
}

bool GLWindow::IsReleased(int key)
{
  return (glfwGetKey(m_window, key) == GLFW_RELEASE);
}