#include "CartRidge.hpp"

#include "mbc/MemBC0.hpp"
#include "mbc/MemBC1.hpp"
#include "mbc/MemBC3.hpp"
#include "mbc/MemBC5.hpp"
#include "Debugger.hpp"

#include <stdexcept>
#include <string>
#include <sstream>
#include <fstream>
#include <iomanip>

#include "../libs/InOutAlgo.hpp"

using namespace std;
using namespace ioa;


const int BANK_SIZE = 0x40*0x100;

CartRidge::CartRidge(Debugger& debugger)
  : m_debugger(debugger),
    m_mbcBPRanges(0),
    m_mbcBP(false)
{
  // Bank 0 and 1 always exist
  m_romBanks.emplace_back(new GBByte[BANK_SIZE]);
  m_romBanks.emplace_back(new GBByte[BANK_SIZE]);

  // command to activate mbc breakpoint
  m_debugger.addCommand("embc", [this](Debugger::ArgList args)
  {
    m_mbcBP = true;
    cout << FG_GREEN << "Activated MBC-write Breakpoint" << C_RESET_FG;

    if(!args.empty())
    {
      if(args.at(1) == "a")
      {
        cout << " (all ranges)";
        m_mbcBPRanges = 0xFF;
      }
      else
      {
        cout << " at ";
        m_mbcBPRanges = 0;

        for(size_t i = 1; i < args.size(); i++)
        {
          auto range = std::stoi(args.at(i), nullptr, 16);
          if(range > 3 || range < 0)
            throw runtime_error("Invalid Range! Allowed from 0 to 3!");

          m_mbcBPRanges |= (1 << range);
          cout << "[" << 2 * range << "000, " << 2 * range + 1 << "FFF], ";
        }
      }
    }

    cout << endl;
    return false;
  });
  m_debugger.addCommand("dmbc", [this](Debugger::ArgList args)
  {
    m_mbcBP = false;
    cout << FG_GREEN << "Deactivated MBC-write Breakpoint!" << C_RESET_FG
         << endl;
    return false;
  });

  m_debugger.addCommand("pmbc", [this](Debugger::ArgList args)
  {
    cout << *this << endl;
    return false;
  });
}


GBByte CartRidge::readByteRom0(GBPtr src)
{
  // rom bank 0 always exists -> return without checking
  return m_romBanks[0][src];
}

GBByte CartRidge::readByteRomN(GBPtr src)
{
  auto romBankNumber = m_mbc->romNumber();

  // check if selected rom exists
  if(romBankNumber >= static_cast<int>(m_romBanks.size()))
  {
    // TODO: tmp
    cout << FG_RED << "ERROR: " << C_RESET_FG << "Internal bank: 0x"
         << hex << romBankNumber << ", Banks available: 0x" << hex 
         << m_romBanks.size() << endl;
    throw runtime_error("Internal banking error: Current ROM Bank number "
                        "is higher than the number of banks!");
  }

  return m_romBanks[romBankNumber][src];
}

void CartRidge::writeByteRom(GBPtr dst, GBByte data)
{
  if(m_mbcBP)
  {
    if((1 << ((dst & 0x6000) >> 13)) & m_mbcBPRanges)
    {
      m_debugger.pause("Write to ROM (changing selected ROM and RAM bank)");
    }
  }
  if(m_debugger.getOutputLevel() >= OL_DEBUG)
  {
    cout << "Write access to external ROM @ " << FG_PURPLE << setfill('0')
         << setw(4) << +dst << C_RESET_FG << " ( " << setw(2) << +data << " )" 
         << endl;
    cout << "Before: " << *this << endl;
  }
  m_mbc->controllWrite(dst, data);
  if(m_debugger.getOutputLevel() >= OL_DEBUG)
  {
    cout << "After: " << *this << endl;
  }
}


std::ostream& operator<<(std::ostream& out, const CartRidge& cr)
{
  cout << "RomBank: " << FG_BLUE << cr.m_mbc->romNumber() << C_RESET_FG
       << ", RamBank: " << FG_BLUE << cr.m_mbc->ramNumber() << C_RESET_FG;
  return out;
}

void CartRidge::readFromFile(const std::string& filename)
{
  // ##### read file ######
  ifstream file("../data/" +  filename, std::ios::binary);

  if(!file)
    throw runtime_error("Could not open file: " +  filename);

  // check size
  file.seekg(0, ios_base::end);
  if(static_cast<int>(file.tellg()) < 2 * BANK_SIZE)
    throw runtime_error("ROM is too small!");
  file.seekg(0, ios_base::beg);

  // read data
  file.read(reinterpret_cast<char*>(m_romBanks[0].get()), BANK_SIZE);
  file.read(reinterpret_cast<char*>(m_romBanks[1].get()), BANK_SIZE);


  // ##### parse header #####
  cout << "------ PARSING ROM HEADER -------" << endl;
  // read title
  m_gameTitle =
    string(reinterpret_cast<char*>(m_romBanks[0].get() + 0x134), 11);

  cout << "Title: " << FG_GREEN << m_gameTitle << C_RESET_FG << endl;

  // read cartridge type, rom size and ram size
  auto cartRidgeType = m_romBanks[0][0x147];
  auto romSize = m_romBanks[0][0x148];
  auto ramSize = m_romBanks[0][0x149];
  setupHardware(cartRidgeType, romSize, ramSize);

  // read remaining data
  for(size_t i = 2; i < m_romBanks.size(); i++)
  {
    file.read(reinterpret_cast<char*>(m_romBanks[i].get()), BANK_SIZE);
  }

  cout << "------ HEADER END -------" << endl << endl;
  
}


void CartRidge::setupHardware(int cartRidgeType, int romSizeByte,
                              int ramSizeByte)
{
  // read version
  int mbcVersion;
  switch(cartRidgeType)
  {
    case 0x00:  // ROM ONLY
      mbcVersion = 0;
      m_mbc = unique_ptr<MemBC0>(new MemBC0());
      break;

    case 0x01:  // MBC1
    case 0x02:  // MBC1 + RAM
    case 0x03:  // MBC1 + RAM + Battery
      mbcVersion = 1;
      m_mbc = unique_ptr<MemBC1>(new MemBC1());
      break;

    case 0x05:  // MBC2
    case 0x06:  // MBC2 + Battery
      mbcVersion = 2;
      break;

    case 0x0F:  // MBC3 + TIMER + Battery
    case 0x10:  // MBC3 + TIMER + RAM + Battery
    case 0x11:  // MBC3
    case 0x12:  // MBC3 + RAM
    case 0x13:  // MBC3 + RAM + Battery
      mbcVersion = 3;
      m_mbc = unique_ptr<MemBC3>(new MemBC3());
      break;

    case 0x19:
    case 0x1A:
    case 0x1B:
    case 0x1C:
    case 0x1D:
    case 0x1E:
      mbcVersion = 5;
      m_mbc = unique_ptr<MemBC5>(new MemBC5());
      break;

    default:
      throw runtime_error("MBC Version not supported!");
  }

  cout << "MBC Version: " << FG_BLUE << mbcVersion << C_RESET_FG << endl;


  // manage rom memory
  int numberRoms;
  if(romSizeByte < 0x52)
    numberRoms = (2 << romSizeByte);
  else
  {
    switch(romSizeByte)
    {
      case 0x52:  numberRoms = 72; break;
      case 0x53:  numberRoms = 80; break;
      case 0x54:  numberRoms = 96; break;
      default:
       throw runtime_error("Invalid ROM size byte!");
    }
  }

  cout << "ROM size: " << FG_BLUE << "0x" << hex
       << numberRoms*(BANK_SIZE / 1024) << C_RESET_FG << " KiB (" << FG_BLUE
       << "0x" << hex << numberRoms << C_RESET_FG << " banks)" << endl;

  for(int i = 2; i < numberRoms; i++)
    m_romBanks.emplace_back(new GBByte[BANK_SIZE]);

  // manage ram size
  int ramSize = 1;
  switch(ramSizeByte)
  {
    case 0x00: ramSize = 0; break;
    case 0x01: ramSize = 2; break;
    case 0x02: ramSize = 8; break;
    case 0x03: ramSize = 32; break;
    default:
      throw runtime_error("Invalid RAM size byte!");
  }
  cout << "RAM size: " << FG_BLUE << hex << "0x" << ramSize << C_RESET_FG
       << " KiB (" << FG_BLUE << "0x" << hex
       << ((ramSize == 0) ? 0 : ((ramSize == 32) ? 4 : 1)) << C_RESET_FG
       << " banks)" << endl;

  m_ram = unique_ptr<GBByte[]>(new GBByte[ramSize*1024]);

  // TODO: check persistance of values
}