#include "Debugger.hpp"
#include "GameBoy.hpp"

#include <iostream>
using namespace std;

#include <boost/algorithm/string.hpp>
using namespace boost;
using namespace boost::algorithm;

#include "../libs/InOutAlgo.hpp"
using namespace ioa;



Debugger::Debugger(GameBoy& gb)
  : m_gb(gb), 
    m_cpu(m_gb.getCPU()),
    m_paused(false),
    m_outputLevel(OL_WARNING)
{
  // add commands
  addDebugCommands();
  resume();

  // by default stop after startup procedure is finished
  m_pcBP.push_back(0x100);
}

void Debugger::check() 
{
  // print current status (after last cpu cycle)
  if(getOutputLevel() >= OL_DEBUG)
  {
    cout << FG_BLUE << m_cpu << C_RESET_FG << endl;
    cout << "-------------------------------" << endl;
  }

  // if we are already paused, we check for user commands
  if(isPaused())
    handleUserInput();

  // check if a new checkpoint is reached
  auto nextIns = m_cpu.getNextIns();
  auto paused = checkForBreakPoint(nextIns);

  // if we reached a breakpoint (this cycle)
  if(paused)
  {
    // print some current information
    cout << FG_BLUE << m_cpu << C_RESET_FG << endl;
    cout << "-------------------------------" << endl;

    pause();

    handleUserInput();
  }

  if(getOutputLevel() >= OL_DEBUG)
  {
    cout << "Executing: " << FG_GREEN << hex << setfill('0') << setw(2)
         << +nextIns << C_RESET_FG << endl;
  }
}

void Debugger::pause(std::string reason)
{
  m_paused = true;
  m_outputLevel = OL_DEBUG;
}

void Debugger::resume()
{
  m_paused = false;
  m_outputLevel = OL_WARNING;
}

bool Debugger::checkForBreakPoint(GBByte nextIns)
{
  // if it is already paused, we ignore checkpoints
  if(isPaused())
    return false;

  // check if any pc-bp applies
  auto pc = m_cpu.getPC();
  for(auto& pcBP : m_pcBP)
  {
    if(pcBP == pc)
    {
      cout << FG_YELLOW << "$$$ Breakpoint reached (PC: " <<  FG_PURPLE 
           << setfill('0') << setw(4) << hex << +pc
           << FG_YELLOW << ") $$$" << C_RESET_FG << endl;
      return true;
    }
  }

  // check if next instruction is any ins-bp
  for(auto& ins : m_insBP)
  {
    if(ins == nextIns)
    {
      cout << FG_YELLOW << "$$$ Breakpoint reached (INS: " <<  FG_PURPLE 
           << setfill('0') << setw(2) << hex << +nextIns << FG_YELLOW
           << ") $$$" << C_RESET_FG << endl;
      return true;
    }
  }
  return false;
}

void Debugger::writeAccess(GBPtr ptr, GBByte data)
{
  for(auto& mem : m_memBP)
  {
    if(mem == ptr)
    {
      cout << FG_YELLOW << "$$$ Breakpoint reached (MEM: write at " << FG_PURPLE
           << setfill('0') << setw(4) << hex << +ptr << " (" << setw(2) << +data
           << ")" << FG_YELLOW << ") $$$" << C_RESET_FG << endl;
      pause();
    }
  }
}

void Debugger::readAccess(GBPtr ptr, GBByte data)
{
  for(auto& mem : m_memBP)
  {
    if(mem == ptr)
    {
      cout << FG_YELLOW << "$$$ Breakpoint reached (MEM: read at " << FG_PURPLE
           << setfill('0') << setw(4) << hex << +ptr << " (" << setw(2) << +data
           << ")" << FG_YELLOW << ") $$$" << C_RESET_FG << endl;
      pause();
    }
  }
}


void Debugger::handleUserInput()
{
  while(true)
  {
    // read input and remove whitespace at beginning and end
    string input;
    getline(cin, input);
    trim(input);

    // if a empty line -> break user input
    if(input.size() == 0)
      break;

    // extract information
    auto firstSpace = input.find(' ');
    auto command = input.substr(0, firstSpace);

    if(m_commands.count(command) == 0)
    {
      cout << "Unknown command (" << FG_RED << command << C_RESET_FG
           << ")" << endl;
    }
    else
    {
      try
      {
        // split input into args and call command
        vector<string> args;
        split(args, input, is_any_of(" "));
        auto ret = m_commands[command](args);

        // decide wether or not to continue user input loop
        if(ret)
          break;
      }
      catch(runtime_error e)
      {
        cout << FG_RED << "EXCEPTION: " << C_RESET_FG
             << "Invalid Input? (what: " << e.what() << ")" << endl;
      }
      catch(...)
      {
        cout << FG_RED << "EXCEPTION: " << C_RESET_FG << "Invalid Input?" << endl;
      }
    }
  }
}

void Debugger::addCommand(const std::string& name, DebugCommand func)
{
  auto& f = m_commands[name];
  if(f)
    std::cout << FG_RED << "Debug command '" << name << "' already taken!" << C_RESET_FG << endl;
  else
    f = func;
}

void Debugger::addDebugCommands()
{
  m_commands["run"] = [this](ArgList) { resume(); return true; };
  m_commands["lj"] = [this](ArgList) { m_cpu.printLastJumps(); return false; };

  m_commands["addpc"] = [this](ArgList args)
  {
    GBPtr pc = std::stoi(args.at(1), nullptr, 16);
    m_pcBP.push_back(pc);
    cout << FG_GREEN << "PCBreakPoint " << hex << setfill('0')
         << setw(4) << pc << " was added!" << C_RESET_FG << endl;
    return false;
  };

  m_commands["delpc"] = [this](ArgList args)
  {
    GBPtr pc = std::stoi(args.at(1), nullptr, 16);
    m_pcBP.erase(remove(m_pcBP.begin(), m_pcBP.end(), pc), m_pcBP.end());
    cout << FG_GREEN << "PCBreakPoint " << hex << setfill('0') 
         << setw(4) << pc << " was deleted!" << C_RESET_FG << endl;
    return false;
  };

  m_commands["addins"] = [this](ArgList args)
  {
    GBByte pc = std::stoi(args.at(1), nullptr, 16);
    m_insBP.push_back(pc);
    cout << FG_GREEN << "INSBreakPoint " << hex << setfill('0')
         << setw(2) << +pc << " was added!" << C_RESET_FG << endl;
    return false;
  };

  m_commands["delins"] = [this](ArgList args)
  {
    GBByte pc = std::stoi(args.at(1), nullptr, 16);
    m_insBP.erase(remove(m_insBP.begin(), m_insBP.end(), pc), m_insBP.end());
    cout << FG_GREEN << "INSBreakPoint " << hex << setfill('0') 
         << setw(2) << +pc << " was deleted!" << C_RESET_FG << endl;
    return false;
  };

  m_commands["addmem"] = [this](ArgList args)
  {
    GBPtr pc = std::stoi(args.at(1), nullptr, 16);
    m_memBP.push_back(pc);
    cout << FG_GREEN << "MEMBreakPoint " << hex << setfill('0')
         << setw(4) << pc << " was added!" << C_RESET_FG << endl;
    return false;
  };

  m_commands["delmem"] = [this](ArgList args)
  {
    GBPtr pc = std::stoi(args.at(1), nullptr, 16);
    m_memBP.erase(remove(m_memBP.begin(), m_memBP.end(), pc), m_memBP.end());
    cout << FG_GREEN << "MEMBreakPoint " << hex << setfill('0') 
         << setw(4) << pc << " was deleted!" << C_RESET_FG << endl;
    return false;
  };
  
  // Add shortcuts
  m_commands["r"] = m_commands["run"];
}
