  #include "GameBoy.hpp"
#include <stdexcept>
#include <thread>
#include <iomanip>
#include <string>
#include <algorithm>

#include "../libs/InOutAlgo.hpp"

using namespace std;
using namespace ioa;

GameBoy::GameBoy()
  : m_interruptEnableRegister(0),
    m_CPU(this),
    m_debugger(*this),
    m_GPU(this),
    m_timer(this),
    m_cartRidge(m_debugger),
    m_workRam(new GBByte[8192]),  // 8 KB working RAM
    m_highRam(new GBByte[127]),   // 127 Byte high RAM
    m_window(160*2, 144*2, "Game Boy Emulator")
{
  memset(&m_registers, 0, sizeof(m_registers));
  m_CPU.addDebugCommands();
  //m_cartRidge.readFromFile("tetris.gb");
  //m_cartRidge.readFromFile("pokemon_yellow.gb");
  //m_cartRidge.readFromFile("kirbys_dream_land.gb");
  //m_cartRidge.readFromFile("wario_land_2.gbc");
  //m_cartRidge.readFromFile("super_mario_land.gb");
  m_cartRidge.readFromFile("test_sound/02-len ctr.gb");

  // Debug commands
  m_debugger.addCommand("highram", [this](Debugger::ArgList args)
  {
    std::cout << std::hex << std::setfill('0');
    for(int i = 0; i < 8; ++i)
    {
      for(int j = 0; j < 16; ++j)
        std::cout << std::setw(2) << +this->readByteImpl(0xFF80 + i*16 + j);
      std::cout << std::endl;
    }
    return false;
  });
}

void GameBoy::writeByte(GBPtr dst, GBByte data)
{  
  if(m_debugger.getOutputLevel() >= OL_DEBUG)
    cout << setfill('0') << "Write access @ " 
         << hex << setw(4) << +dst 
         << " (" << setw(2) << +data << ")" << endl;


  try
  {
    this->writeByteImpl(dst, data);
  }
  catch(...)
  {
    cout << FG_RED << "Exception while read access @ " << hex 
         << setfill('0') << setw(4) << +dst
         << " (" << setw(2) << +data << ")" << C_RESET_FG << endl;
    m_CPU.printRegisters();
    throw;
  }
}

void GameBoy::writeByteImpl(GBPtr dst, GBByte data)
{
  // notify debugger 
  m_debugger.writeAccess(dst, data);

  if(dst < 0x4000)
  {
    // Note: I'm not sure what happens if the internal startup rom is still
    // enabled and you try to write to the memory < FFh. But it won't happen...
    m_cartRidge.writeByteRom(dst, data);
  }
  else if(dst < 0x8000)
  {
    m_cartRidge.writeByteRom(dst, data);
  }
  else if(dst < 0xA000)  // 8KB VRAM [8000h, 9FFFFh]
  {
    m_GPU.getVRam()[dst - 0x8000] = data;
  }
  else if(dst < 0xC000)  // 8KB external RAM [A000h, BFFFh]
  {
    //throw runtime_error("Address range not implemented!");
  }
  else if(dst < 0xE000)  // 8KB work RAM [C000h, DFFFh]
  {
    m_workRam[dst - 0xC000] = data;
  }
  else if(dst < 0xFE00)  // echo range for work RAM [E000h, FDFFh]
  {
    m_workRam[dst - 0xE000] = data;
  }
  else if(dst < 0xFEA0)  // sprite attribute table (OAM) [FE00h, FE9F]
  {
    m_GPU.getSpriteAttributes()[dst - 0xFE00] = data;
  }
  else if(dst < 0xFF00)  // not usable [FEA0h, FEFFh]
  {
    /*cout << FG_RED << "Write to Address range [FEA0h, FEFFh] (not usable)! @"
         << hex << dst << ", PC: " << m_CPU.getPC() << C_RESET_FG << endl;*/
    //throw runtime_error("Address range [FEA0h, FEFFh] not usable!");
  }
  else if(dst < 0xFF80)  // IO Ports [FF00h, FF7F]
  {
    this->writeRegister(dst - 0xFF00, data);
  }
  else if(dst != 0xFFFF)  // High RAM [FF80, FFFE]
  {
    m_highRam[dst - 0xFF80] = data;
  }
  else  // Interrupt Enable Register @ FFFFh
  {
    m_interruptEnableRegister = data;
  } 
}

void GameBoy::writeRegister(GBByte dst, GBByte data)
{
  // Ignore writes if sound circuit is disabled
  HardwareRegisters* r = &m_registers;
  if(dst >= 0x10 && dst <= 0x25 && (~r->NR52 & 0x80))
    return;

  // Handle other cases
  GBByte* p = reinterpret_cast<GBByte*>(r) + dst;
  switch(dst)
  {
    case 0x04: // Writing resets to 0
      r->DIV = 0;
      break;
    case 0x26: // Disabling sound circuit resets registers
      if(~data & 0x80)
        memset(&r->NR10, 0, 22);
      *p = (*p & ~regWriteMask[dst]) | (data & regWriteMask[dst]);
      break;
    case 0x44: // Writing resets to 0
      r->LY = 0;
      break;
    case 0x46: // DMA transfer
    {
      GBPtr src = data * 0x100;
      for(int i = 0; i < 160; ++i)
        this->writeByte(0xFE00 + i, this->readByte(src + i));
      break;
    }
    default:
      *p = (*p & ~regWriteMask[dst]) | (data & regWriteMask[dst]);
  }
}

GBByte GameBoy::readByte(GBPtr ptr)
{  
  if(m_debugger.getOutputLevel() >= OL_DEBUG)
    cout << setfill('0') << "Read access @ " 
         << hex << setw(4) << +ptr;
  
  GBByte out;
  try
  {
    out = this->readByteImpl(ptr);
    m_debugger.readAccess(ptr, out);
  }
  catch(...)
  {
    cout << FG_RED << "Exception while read access @ " << hex 
         << setfill('0') << setw(4) << +ptr << C_RESET_FG << endl;
    m_CPU.printRegisters();
    throw;
  }
  if(m_debugger.getOutputLevel() >= OL_DEBUG)
    cout << " (" << setw(2) << +out << ")" << endl;
  return out;
}

GBByte GameBoy::readByteImpl(GBPtr ptr)
{
  if(ptr < 0x4000)
  {
    if(m_registers.ROMD == 0 && ptr <= 0xFF)
    {
      return m_CPU.getInternalRom()[ptr];
    }
    return m_cartRidge.readByteRom0(ptr);
  }
  if(ptr < 0x8000)
  {
    return m_cartRidge.readByteRomN(ptr - 0x4000);
  }
  if(ptr < 0xA000)  // 8KB VRAM [8000h, 9FFFFh]
  {
    return m_GPU.getVRam()[ptr - 0x8000];
  }
  if(ptr < 0xC000)  // 8KB external RAM [A000h, BFFFh]
  {
    return 0;
    //throw runtime_error("Address range not implemented!");
  }
  if(ptr < 0xE000)  // 8KB work RAM [C000h, DFFFh]
  {
    return m_workRam[ptr - 0xC000];
  }
  if(ptr < 0xFE00)  // echo range for work RAM [E000h, FDFFh]
  {
    return m_workRam[ptr - 0xE000];
  }
  if(ptr < 0xFEA0)  // sprite attribute table (OAM) [FE00h, FE9F]
  {
    return m_GPU.getSpriteAttributes()[ptr - 0xFE00];
  }
  if(ptr < 0xFF00)  // not usable [FEA0h, FEFFh]
  {
    /*cout << FG_RED << "Read from Address range [FEA0h, FEFFh] (not usable)! @"
         << hex << ptr << C_RESET_FG << endl;*/
    return 0;
    //throw runtime_error("Address range [FEA0h, FEFFh] not usable!");
  }
  if(ptr < 0xFF80)  // IO Ports [FF00h, FF7F]
  {
    return this->readRegister(ptr - 0xFF00);
  }
  if(ptr != 0xFFFF)  // High RAM [FF80, FFFE]
  {
    return m_highRam[ptr - 0xFF80];
  }
  
  // Interrupt Enable Register @ FFFFh
  return m_interruptEnableRegister;
}

GBByte GameBoy::readRegister(GBByte ptr)
{
  HardwareRegisters* r = &m_registers;
  switch(ptr)
  {
    case 0x00:
      if(~r->JOYP & 0x10) // Direction keys
      {
        GBByte keys = 0;
        keys |= (m_window.IsReleased(GLFW_KEY_RIGHT) << 0);
        keys |= (m_window.IsReleased(GLFW_KEY_LEFT)  << 1);
        keys |= (m_window.IsReleased(GLFW_KEY_UP)    << 2);
        keys |= (m_window.IsReleased(GLFW_KEY_DOWN)  << 3);
        return keys | (r->JOYP & 0xF0);
      }
      if(~r->JOYP & 0x20) // Button keys
      {
        GBByte keys = 0;
        keys |= (m_window.IsReleased(GLFW_KEY_A)         << 0); // Button A
        keys |= (m_window.IsReleased(GLFW_KEY_S)         << 1); // Button B
        keys |= (m_window.IsReleased(GLFW_KEY_BACKSPACE) << 2); // Select
        keys |= (m_window.IsReleased(GLFW_KEY_ENTER)     << 3); // Start
        return keys | (r->JOYP & 0xF0);
      }
      return 0x0F | (r->JOYP & 0xF0); // Return all keys released
    default:
      return reinterpret_cast<GBByte*>(&m_registers)[ptr] | regReadMask[ptr];
  }
}

std::ostream& operator<<(std::ostream& out, const GameBoy& gb) 
{
  out << "##### CartRidge #####" << endl;
  out << gb.m_cartRidge;

  return out;
}

void addBPRange(vector<GBPtr>& v, GBPtr from, GBPtr to)
{
  while(from <= to)
    v.push_back(from++);
}

void GameBoy::runDevice()
{
  using namespace std::chrono;
  try
  {
    while(m_window.Run())
    {

      // check for interrupts first because it may change the PC
      m_CPU.checkInterrupts();

      // do debug stuff
      m_debugger.check();

      // Cycle CPU
      MachineCycle cycles = m_CPU.cycle();

      // Cycle other hardware
      for(uint64_t i = 0; i < cycles.count(); ++i)
      {
        m_GPU.cycle();
        m_timer.cylce();
      }
    }
  }
  catch(...)
  {
    cout << FG_RED << "Exception! Last 20 Jumps: " 
         << C_RESET_FG << endl;
    m_CPU.printLastJumps();
    m_CPU.printRegisters();
    throw;
  }
}
