#include "GraphicProcessor.hpp"
#include "GameBoy.hpp"
#include <algorithm>

GraphicProcessor::GraphicProcessor(GameBoy* gb)
  : m_gb(gb),
    m_vRam(new VRAM),
    m_sprites(new SpriteAttribute[40]),
    m_cycleCount(0),
    m_screen(new uint8_t[160*144]),
    m_virtualScreen(new uint8_t[256])
{
}

GraphicProcessor::~GraphicProcessor()
{
}

void GraphicProcessor::cycle()
{
  HardwareRegisters* r = m_gb->getRegisters();
  if((r->LCDC >> 7) == 0) // LCD disabled
    return;

  // Set LCD Mode Flag (Bit 0-1)
  if(r->LY >= 144) // Mode 1 - VBlank
  {
    r->STAT |=  (1 << 0);
    r->STAT &= ~(1 << 1);
    if(r->LY == 144)
    {
      r->IF |= (1 << INTERRUPT_VBLANK);
      if((r->STAT >> 4) & 0x01)
        r->IF |= (1 << INTERRUPT_LCDC_STAT);
    }
  }
  else if(m_cycleCount == 0) // Mode 2 - Read OAM (20 cycles)
  {
    r->STAT &= ~(1 << 0);
    r->STAT |=  (1 << 1);
    this->buildSpriteTable();
    if((r->STAT >> 5) & 0x01)
      r->IF |= (1 << INTERRUPT_LCDC_STAT);
  }
  else if(m_cycleCount == 20) // Mode 3 - Read OAM and VRAM (43 cycles)
  {
    r->STAT |= (1 << 0);
    r->STAT |= (1 << 1);
    this->drawLine();
  }
  else if(m_cycleCount == 20+43) // Mode 0 - HBlank (51 cycles)
  {
    r->STAT &= ~(1 << 0);
    r->STAT &= ~(1 << 1);
    if((r->STAT >> 3) & 0x01)
      r->IF |= (1 << INTERRUPT_LCDC_STAT);
  }

  // Count cycles (114 cycles per line)
  if(++m_cycleCount == 114)
  {
    m_cycleCount = 0;

    // Goto next line
    if(++r->LY == 154)
    {
      r->LY = 0;
      this->draw();
    }

    // Update Coincidence Flag (Bit 2)
    if(r->LY == r->LYC)
    {
      r->STAT |= (1 << 2);
      if((r->STAT >> 6) & 0x01)
        r->IF |= (1 << INTERRUPT_LCDC_STAT);
    }
    else
    {
      r->STAT &= ~(1 << 2);
    }
  }
}

void GraphicProcessor::draw()
{
  GLWindow* wnd = m_gb->getWindow();
  wnd->Clear();
  wnd->DrawTexture(m_screen.get(), 160, 144);
  wnd->Present();
}

void GraphicProcessor::drawLine()
{
  HardwareRegisters* r = m_gb->getRegisters();
  GBByte clr;

  // Render background
  if(r->LCDC & 1)
  {
    GBByte* bgMap = m_vRam->backgroundMap[(r->LCDC >> 3) & 1].data;
    uint8_t mask = ((r->LCDC >> 4) & 1) ? 0 : 0x80;
    Tile* tiles = m_vRam->tiles + mask;

    int y = (r->LY + r->SCY) % 256;
    for(int x = 0; x < 32; ++x)
    {
      // Unpack tile
      int sx = x*8 + (256 - r->SCX);
      const Tile& t = tiles[bgMap[(y/8)*32 + x] ^ mask];
      for(int col = 0; col < 8; ++col)
      {
        clr =  ((t.data[(y % 8)*2]   >> (7-col)) & 1) ? 1 : 0;
        clr += ((t.data[(y % 8)*2+1] >> (7-col)) & 1) ? 2 : 0;
        m_virtualScreen[(sx+col) % 256] = (r->BGP >> clr*2) & 0x03;
      }
    }
  }

  // Render sprites
  if((r->LCDC >> 1) & 1)
  {
    int height = ((r->LCDC >> 2) & 1) ? 16 : 8;
    for(auto s : m_spritePriority)
    {
      // Load sprite
      const SpriteAttribute& sa = m_sprites[s];
      GBByte palette = ((sa.flags >> 4) & 1) ? r->OBP1 : r->OBP0;
      bool flipX = ((sa.flags >> 5) & 1);
      bool flipY = ((sa.flags >> 6) & 1);
      int x = sa.xPos - 8;
      int y = sa.yPos - 16;

      // Unpack tile
      int row = flipY ? (height - r->LY + y) : (r->LY - y);
      const Tile& t = m_vRam->tiles[sa.tileNum];
      for(int col = std::max(0, -x); col < 8; ++col)
      {
        int shift = flipX ? col : (7-col);
        clr =  ((t.data[row*2]   >> shift) & 1) ? 1 : 0;
        clr += ((t.data[row*2+1] >> shift) & 1) ? 2 : 0;
        if(clr != 0) // Transparency
          m_virtualScreen[x + col] = (palette >> clr*2) & 0x03;
      }
    }
  }

  // Copy to screen
  uint8_t colors[4] = {255, 170, 85, 0};
  for(int x = 0; x < 160; ++x)
    m_screen[r->LY*160 + x] = colors[m_virtualScreen[x]];
}

void GraphicProcessor::buildSpriteTable()
{
  m_spritePriority.clear();
  HardwareRegisters* r = m_gb->getRegisters();

  // Get all visible sprites
  int height = ((r->LCDC >> 2) & 1) ? 16 : 8;
  for(int i = 0; i < 40; ++i)
  {
    int y = m_sprites[i].yPos - 16;
    if(y+height > r->LY && r->LY >= y)
      m_spritePriority.push_back(i);
  }

  // Sort by priority
  std::sort(m_spritePriority.begin(), m_spritePriority.end(),
    [this](uint8_t r, uint8_t l)
    {
      if(m_sprites[r].xPos < m_sprites[l].xPos)
        return true;
      else if(m_sprites[r].xPos > m_sprites[l].xPos)
        return false;
      return (r < l);
    }
  );
}
