#include "MemoryBankController.hpp"

#include <stdexcept>

#include "../libs/InOutAlgo.hpp"

using namespace std;
using namespace ioa;


void MemoryBankController::create(int version)
{
  m_mbcVersion = version;
}

int MemoryBankController::currentRomBankNumber() const
{
  switch(m_mbcVersion)
  {
    case 0:
      return 1;
    case 1:
    {
      // extract either 5 or 7 bits from the m_controlRegister 
      int num =
          m_romModeEnabled ? m_controlRegister & 0x7F : m_controlRegister & 0x1F;
      
      // correct value if the first 5 bits are 0
      if((num & 0x1F) == 0)
        num++;

      return num;
    }
    case 2:
    {
      // extract exact 4 bits
      int num = m_controlRegister & 0x0F;

      // just correct if the value is 0
      if(num == 0)
        num++;

      return num;
    }
    case 3:
    {
      // extract exact 7 bits
      int num = m_controlRegister & 0x7F;

      // just correct if the value is 0
      if(num == 0)
        num++;

      return num;
    }
    case 5:
    {
      return m_romNumber;
    }
  }
  throw runtime_error("MemoryBankController was not initialized "
                      "(invalid version number)!");
}



int MemoryBankController::currentRamBankNumber() const
{
  switch(m_mbcVersion)
  {
    case 0:
      return 0;
    case 1:
      return m_romModeEnabled ? 0 : (m_controlRegister & 0x60) >> 5;
    case 2:
      break;
    case 3:
      return m_controlRegisterB & 0x0F;
    case 5:
      return m_ramNumber;
  }
  throw runtime_error("MemoryBankController was not initialized "
                      "(invalid version number)!");
}


void MemoryBankController::writeControllA(GBByte data)
{
  m_controlRegister = (m_controlRegister & 0x7F) | (((data & 0x0F) == 0x0A) << 7);
}


// TODO: "The least significant bit of the upper address byte must 
// be one to select a ROM bank" http://nocash.emubase.de/pandocs.htm
// same for ram enable... bullshit.
// TODO: This code fits nicer into the MemoryBankController class
void MemoryBankController::writeControllB(GBByte data, GBPtr ptr)
{
  // writing in this memory range changes the selected ROM bank
  switch(m_mbcVersion)
  {
  case 1:
    // writing first 5 bits to register
    m_controlRegister = (m_controlRegister & 0xE0) | (data & 0x1F);
    break;
  case 2:
    // writing 4 bits into register
    m_controlRegister = (m_controlRegister & 0xF0) | (data & 0x0F);
    break;
  case 3:
    // writing 7 bits into register
    m_controlRegister = (m_controlRegister & 0x80) | (data & 0x7F);
    break;
  case 5:
    if(ptr < 3000)
      m_romNumber = (m_romNumber & 0xFF00) | data;
    else
      m_romNumber = (m_romNumber & 0xFEFF) | ((data & 0x01) << 8);
    break;
  }
}

void MemoryBankController::writeControllC(GBByte data)
{
  switch(m_mbcVersion)
  {
  case 1:
    // writing two bits, later interpreted as RAM bank number or as part
    // of the ROM bank number
    m_controlRegister = (m_controlRegister & 0x9F) | ((data << 5) & 0x60);
    break;
  case 3:
    // writing 4 bits (RAM bank or RTC)
    m_controlRegisterB = (m_controlRegisterB & 0xF0) | (data & 0x0F);
    break;
  case 5:
    m_ramNumber = data & 0x0F;
  }
}

void MemoryBankController::writeControllD(GBByte data)
{
  switch(m_mbcVersion)
  {
  case 1:
    // enabling ROM mode if data is 0
    m_romModeEnabled = (data == 0);
    break;
  case 3:
    // TODO: Latch Clock Data
    break;
  }
}


std::ostream& operator<<(std::ostream& out, const MemoryBankController& mbc)
{
  out << "ROM/RAM-Mode: " << FG_BLUE << (mbc.m_romModeEnabled ? "ROM" : "RAM")
      << C_RESET_FG << "\tRAM-Enable: " << FG_BLUE
      << ((mbc.m_controlRegister & 0x80) == 0x80) << C_RESET_FG << endl;
  out << "ROM Bank: " << hex << FG_BLUE << mbc.currentRomBankNumber()
      << C_RESET_FG << "\tRAM Bank: " << FG_BLUE << hex
      << mbc.currentRamBankNumber() << C_RESET_FG << endl;

  return out;
}