#include "Processor.hpp"

#include "GameBoy.hpp"
#include "Instructions.hpp"
#include "../libs/InOutAlgo.hpp"

#include <boost/algorithm/string.hpp>
#include <cstring>
#include <fstream>
#include <stdexcept>
#include <iostream>
#include <iomanip>

using namespace std;
using namespace ioa;

uint8_t instructionDuration[256]
  //  0  1  2  3  4  5  6  7  8  9  A  B  C  D  E  F
  = { 1, 3, 2, 2, 1, 1, 2, 1, 5, 2, 2, 2, 1, 1, 2, 1, // 0
      1, 3, 2, 2, 1, 1, 2, 1, 3, 2, 2, 2, 1, 1, 2, 1, // 1
      2, 3, 2, 2, 1, 1, 2, 1, 2, 2, 2, 2, 1, 1, 2, 1, // 2
      2, 3, 2, 2, 3, 3, 3, 1, 2, 2, 2, 2, 1, 1, 2, 1, // 3
      1, 1, 1, 1, 1, 1, 2, 1, 1, 1, 1, 1, 1, 1, 2, 1, // 4
      1, 1, 1, 1, 1, 1, 2, 1, 1, 1, 1, 1, 1, 1, 2, 1, // 5
      1, 1, 1, 1, 1, 1, 2, 1, 1, 1, 1, 1, 1, 1, 2, 1, // 6
      2, 2, 2, 2, 2, 2, 1, 2, 1, 1, 1, 1, 1, 1, 2, 1, // 7
      1, 1, 1, 1, 1, 1, 2, 1, 1, 1, 1, 1, 1, 1, 2, 1, // 8
      1, 1, 1, 1, 1, 1, 2, 1, 1, 1, 1, 1, 1, 1, 2, 1, // 9
      1, 1, 1, 1, 1, 1, 2, 1, 1, 1, 1, 1, 1, 1, 2, 1, // A
      1, 1, 1, 1, 1, 1, 2, 1, 1, 1, 1, 1, 1, 1, 2, 1, // B
      2, 3, 3, 4, 3, 4, 2, 4, 2, 4, 3, 1, 3, 6, 2, 4, // C
      2, 3, 3, 0, 3, 4, 2, 4, 2, 4, 3, 0, 3, 0, 2, 4, // D
      3, 3, 2, 0, 0, 4, 2, 4, 4, 1, 4, 0, 0, 0, 2, 4, // E
      3, 3, 2, 1, 0, 4, 2, 4, 3, 2, 4, 1, 0, 0, 2, 4};// F

uint8_t instructionDurationCB[16]
  //  0  1  2  3  4  5  6  7  8  9  A  B  C  D  E  F
  = { 2, 2, 2, 2, 2, 2, 4, 2, 2, 2, 2, 2, 2, 2, 4, 2 };


Processor::Processor(GameBoy* gb)
  : m_gb(gb),
    m_halt(false),
    m_stop(false),
    m_startUpRom(new GBByte[256])  // 256 byte startup rom
{
  // reset all register
  memset(&m_register, 0, sizeof(m_register));

  // load startup sequence in internal rom
  ifstream file("../data/DMG_ROM.bin", std::ios::binary);

  if(!file)
    throw runtime_error("Could not open file with startup sequence!");

  // check size
  file.seekg(0, ios_base::end);
  if(static_cast<int>(file.tellg()) != 256)
    throw runtime_error("File with startup has an invalid length!");
  file.seekg(0, ios_base::beg);

  // TODO: maybe calculate checksum of startup rom

  file.read(reinterpret_cast<char*>(m_startUpRom.get()), 256);
}


GBByte Processor::getNextIns() const
{
  // read without output
  return m_gb->readByte(m_register.PC);
  //return m_gb->readByteImpl(m_register.PC);
}

void Processor::checkInterrupts()
{
  HardwareRegisters* io = m_gb->getRegisters();
  GBByte interrupt = io->IF & m_gb->readByte(0xFFFF);
  if(interrupt && m_interruptMasterEnableRegister)
  {
    for(int i = 0; i < 5; ++i)
    {
      GBByte interruptBit = (1 << i);
      if(interrupt & interruptBit)
      {
        // Reset flags
        m_interruptMasterEnableRegister = 0;
        io->IF &= ~interruptBit;

        // Call to interrupt
        auto& r = m_register;
        r.SP -= 2;
        m_gb->writeWord(r.SP, r.PC);
        addJump(r.PC);
        r.PC = 0x40 + i*8;
        return;
      }
    }
  }
}

void Processor::addDebugCommands()
{
  auto f = [this](Debugger::ArgList args)
  {
    // Parameters
    std::string reg = boost::algorithm::to_upper_copy(args.at(1));
    GBWord val = std::stoi(args.at(2), nullptr, 16);

    // Change register
    auto& r = m_register;
    if(reg.size() == 1)
    {
      // Single character registers
      switch(reg[0])
      {
        case 'A':
          r.A = val & 0xFF;
          break;
        case 'B':
          r.B = val & 0xFF;
          break;
        case 'C':
          r.C = val & 0xFF;
          break;
        case 'D':
          r.D = val & 0xFF;
          break;
        case 'E':
          r.E = val & 0xFF;
          break;
        case 'F':
          r.F = val & 0xFF;
          break;
        case 'H':
          r.H = val & 0xFF;
          break;
        case 'L':
          r.L = val & 0xFF;
          break;
        default:
          std::cout << FG_RED << "Unknown register '" << reg << "'. Nothing changed."
                    << C_RESET_FG << endl;
      }
    }
    else if(reg == "AF")
    {
      r.AF = val;
    }
    else if(reg == "BC")
    {
      r.BC = val;
    }
    else if(reg == "DE")
    {
      r.DE = val;
    }
    else if(reg == "HL")
    {
      r.HL = val;
    }
    else if(reg == "SP")
    {
      r.SP = val;
    }
    else if(reg == "PC")
    {
      r.PC = val;
    }
    else
    {
      std::cout << FG_RED << "Unknown register '" << reg << "'. Nothing changed."
                << C_RESET_FG << endl;
    }
    this->printRegisters();
    return false;
  };

  // Set command
  Debugger* d = m_gb->getDebugger();
  d->addCommand("set", f);
  d->addCommand("s", f);
}

MachineCycle Processor::cycle()
{
  // Check for HALT/STOP
  if(m_halt || m_stop)
  {
    GBByte IF = m_gb->getRegisters()->IF;
    if(IF & (1 << INTERRUPT_JOYPAD))
      m_stop = false;
    if(IF)
      m_halt = false;

    if(m_halt || m_stop)
      return MachineCycle(1);
  }

  /*// Temp!!!
  if(r.PC == 0x03)
    cout << "---------- Zero RAM..." << endl;
  if(r.PC == 0x0c)
    cout << "---------- Setup Audio..." << endl;
  if(r.PC == 0x21)
    cout << "---------- Load Logo..." << endl;
  if(r.PC == 0x40)
    cout << "---------- Setup Tilemap..." << endl;
  if(r.PC == 0x56)
    cout << "---------- Start Logo Scrolling..." << endl;
  if(r.PC == 0x91)
    cout << "---------- Showing Logo a little longer..." << endl;
  if(r.PC == 0xE0)
    cout << "---------- Comparing Logo Data..." << endl;
  if(r.PC == 0xF1)
    cout << "---------- Calculating Checksum..." << endl;*/

  auto& r = m_register;
  auto ins = m_gb->readByte(r.PC++);
  int dur = instructionDuration[ins];

  // Temporary variables to store intermediate results
  GBByte b;

  // Exectute instruction
  switch(ins)
  {
    case 0x00:  // NOP
      break;

    // ### LD r, n  (8bit immediate)
    case 0x06:  // LD B, n
      r.B = m_gb->readByte(r.PC++);
      break;
    case 0x0E:  // LD C, n
      r.C = m_gb->readByte(r.PC++);
      break;
    case 0x16:  // LD D, n
      r.D = m_gb->readByte(r.PC++);
      break;
    case 0x1E:  // LD E, n
      r.E = m_gb->readByte(r.PC++);
      break;
    case 0x26:  // LD H, n
      r.H = m_gb->readByte(r.PC++);
      break;
    case 0x2E:  // LD L, n
      r.L = m_gb->readByte(r.PC++);
      break;


    // ### LD r, r  (8 bit register)
    case 0x7F:  // LD A, A
      r.A = r.A;
      break;

    case 0x40:  // LD B, B
      r.B = r.B;
      break;
    case 0x41:  // LD B, C
      r.B = r.C;
      break;
    case 0x42:  // LD B, D
      r.B = r.D;
      break;
    case 0x43:  // LD B, E
      r.B = r.E;
      break;
    case 0x44:  // LD B, H
      r.B = r.H;
      break;
    case 0x45:  // LD B, L
      r.B = r.L;
      break;
    case 0x46:  // LD B, [HL]
      r.B = m_gb->readByte(r.HL);
      break;

    case 0x48:  // LD C, B
      r.C = r.B;
      break;
    case 0x49:  // LD C, C
      r.C = r.C;
      break;
    case 0x4A:  // LD C, D
      r.C = r.D;
      break;
    case 0x4B:  // LD C, E
      r.C = r.E;
      break;
    case 0x4C:  // LD C, H
      r.C = r.H;
      break;
    case 0x4D:  // LD C, L
      r.C = r.L;
      break;
    case 0x4E:  // LD C, [HL]
      r.C = m_gb->readByte(r.HL);
      break;

    case 0x50:  // LD D, B
      r.D = r.B;
      break;
    case 0x51:  // LD D, C
      r.D = r.C;
      break;
    case 0x52:  // LD D, D
      r.D = r.D;
      break;
    case 0x53:  // LD D, E
      r.D = r.E;
      break;
    case 0x54:  // LD D, H
      r.D = r.H;
      break;
    case 0x55:  // LD D, L
      r.D = r.L;
      break;
    case 0x56:  // LD D, [HL]
      r.D = m_gb->readByte(r.HL);
      break;

    case 0x58:  // LD E, B
      r.E = r.B;
      break;
    case 0x59:  // LD E, C
      r.E = r.C;
      break;
    case 0x5A:  // LD E, D
      r.E = r.D;
      break;
    case 0x5B:  // LD E, E
      r.E = r.E;
      break;
    case 0x5C:  // LD E, H
      r.E = r.H;
      break;
    case 0x5D:  // LD E, L
      r.E = r.L;
      break;
    case 0x5E:  // LD E, [HL]
      r.E = m_gb->readByte(r.HL);
      break;

    case 0x60:  // LD H, B
      r.H = r.B;
      break;
    case 0x61:  // LD H, C
      r.H = r.C;
      break;
    case 0x62:  // LD H, D
      r.H = r.D;
      break;
    case 0x63:  // LD H, E
      r.H = r.E;
      break;
    case 0x64:  // LD H, H
      r.H = r.H;
      break;
    case 0x65:  // LD H, L
      r.H = r.L;
      break;
    case 0x66:  // LD H, [HL]
      r.H = m_gb->readByte(r.HL);
      break;

    case 0x68:  // LD L, B
      r.L = r.B;
      break;
    case 0x69:  // LD L, C
      r.L = r.C;
      break;
    case 0x6A:  // LD L, D
      r.L = r.D;
      break;
    case 0x6B:  // LD L, E
      r.L = r.E;
      break;
    case 0x6C:  // LD L, H
      r.L = r.H;
      break;
    case 0x6D:  // LD L, L
      r.L = r.L;
      break;
    case 0x6E:  // LD H, [HL]
      r.L = m_gb->readByte(r.HL);
      break;

    case 0x70:  // LD [HL], B
      m_gb->writeByte(r.HL, r.B);
      break;
    case 0x71:  // LD [HL], C
      m_gb->writeByte(r.HL, r.C);
      break;
    case 0x72:  // LD [HL], D
      m_gb->writeByte(r.HL, r.D);
      break;
    case 0x73:  // LD [HL], E
      m_gb->writeByte(r.HL, r.E);
      break;
    case 0x74:  // LD [HL], H
      m_gb->writeByte(r.HL, r.H);
      break;
    case 0x75:  // LD [HL], L
      m_gb->writeByte(r.HL, r.L);
      break;

    case 0x36:  // LD [HL], n
      m_gb->writeByte(r.HL, m_gb->readByte(r.PC++));
      break;
    // ...


    // ### LD A, n (load into A)
    case 0x78:  // LD A, B
      r.A = r.B;
      break;
    case 0x79:  // LD A, C
      r.A = r.C;
      break;
    case 0x7A:  // LD A, D
      r.A = r.D;
      break;
    case 0x7B:  // LD A, E
      r.A = r.E;
      break;
    case 0x7C:  // LD A, H
      r.A = r.H;
      break;
    case 0x7D:  // LD A, L
      r.A = r.L;
      break;
    case 0x0A:  // LD A, [BC]
      r.A = m_gb->readByte(r.BC);
      break;
    case 0x1A:  // LD A, [DE]
      r.A = m_gb->readByte(r.DE);
      break;
    case 0x7E:  // LD A, [HL]
      r.A = m_gb->readByte(r.HL);
      break;
    case 0xFA:  // LD A, [nn]
      r.A = m_gb->readByte(m_gb->readWord(r.PC++));
      r.PC++;
      break;
    case 0x3E:  // LD A, #v
      r.A = m_gb->readByte(r.PC++);
      break;


    // ### LD n, A (load from A)
    case 0x47:  // LD B, A
      r.B = r.A;
      break;
    case 0x4F:  // LD C, A
      r.C = r.A;
      break;
    case 0x57:  // LD D, A
      r.D = r.A;
      break;
    case 0x5F:  // LD E, A
      r.E = r.A;
      break;
    case 0x67:  // LD H, A
      r.H = r.A;
      break;
    case 0x6F:  // LD L, A
      r.L = r.A;
      break;
    case 0x02:  // LD [BC], A
      m_gb->writeByte(r.BC, r.A);
      break;
    case 0x12:  // LD [DE], A
      m_gb->writeByte(r.DE, r.A);
      break;
    case 0x77:  // LD [HL], A
      m_gb->writeByte(r.HL, r.A);
      break;
    case 0xEA:  // LD [nn], A
      m_gb->writeByte(m_gb->readWord(r.PC++), r.A);
      r.PC++;
      break;


    // ### LD A + IO-Port
    case 0xF2:  // LD A, [FF00h + C]
      r.A = m_gb->readByte(0xFF00 + r.C);
      break;
    case 0xE2:  // LD [FF00h + C], A
      m_gb->writeByte(0xFF00 + r.C, r.A);
      break;
    case 0xF0:  // LD A, [FF00h + n]
      r.A = m_gb->readByte(0xFF00 + m_gb->readByte(r.PC++));
      break;
    case 0xE0:  // LD [FF00h + n], A
      m_gb->writeByte(0xFF00 + m_gb->readByte(r.PC++), r.A);
      break;


    // ### LD with decrement/increment from HL
    case 0x3A:  // LDD A, [HL]
      r.A = m_gb->readByte(r.HL--);
      break;
    case 0x32:  // LDD [HL], A
      m_gb->writeByte(r.HL--, r.A);
      break;
    case 0x2A:  // LDI A, [HL]
      r.A = m_gb->readByte(r.HL++);
      break;
    case 0x22:  // LDI [HL], A
      m_gb->writeByte(r.HL++, r.A);
      break;


    // ### LD rr, nn (16bit immediate)
    case 0x01:  // LD BC, nn
      r.BC = m_gb->readWord(r.PC);
      r.PC += 2;
      break;
    case 0x11:  // LD DE, nn
      r.DE = m_gb->readWord(r.PC);
      r.PC += 2;
      break;
    case 0x21:  // LD HL, nn
      r.HL = m_gb->readWord(r.PC);
      r.PC += 2;
      break;
    case 0x31:  // LD SP, nn
      r.SP = m_gb->readWord(r.PC);
      r.PC += 2;
      break;
    case 0x08:  // LD (nn), SP
      m_gb->writeWord(m_gb->readWord(r.PC), r.SP);
      r.PC += 2;
      break;
    case 0xF9:  // LD SP, HL
      r.SP = r.HL;
      break;
    case 0xF8:  // LD HL, SP+r8
    {
      int v = static_cast<int8_t>(m_gb->readByte(r.PC++));
      if(v >= 0)
        r.F = (((r.SP & 0x0F) + v > 0x0F) << 5) | (((r.SP & 0xFF) + v > 0xFF) << 4);
      else
        r.F = (((r.SP & 0x0F) >= (-v & 0x0F)) << 5) | (((r.SP & 0xFF) >= -v) << 4);
      r.HL = r.SP + v;
      break;
    }


    // ### INC nn (++nn)
    case 0x03: ++r.BC; break;
    case 0x13: ++r.DE; break;
    case 0x23: ++r.HL; break;
    case 0x33: ++r.SP; break;

    // ### DEC nn (--nn)
    case 0x0B: --r.BC; break;
    case 0x1B: --r.DE; break;
    case 0x2B: --r.HL; break;
    case 0x3B: --r.SP; break;


    // ### INC n (++n, F=Z0H-)
    case 0x3C: INC(r.A) break;
    case 0x04: INC(r.B) break;
    case 0x0C: INC(r.C) break;
    case 0x14: INC(r.D) break;
    case 0x1C: INC(r.E) break;
    case 0x24: INC(r.H) break;
    case 0x2C: INC(r.L) break;
    case 0x34:  // INC [HL]
    {
      GBByte n = m_gb->readByte(r.HL);
      INC(n);
      m_gb->writeByte(r.HL, n);
      break;
    }


    // ### DEC n (--n, F=Z1H-)
    case 0x3D: DEC(r.A) break;
    case 0x05: DEC(r.B) break;
    case 0x0D: DEC(r.C) break;
    case 0x15: DEC(r.D) break;
    case 0x1D: DEC(r.E) break;
    case 0x25: DEC(r.H) break;
    case 0x2D: DEC(r.L) break;
    case 0x35:  // DEC [HL]
    {
      GBByte n = m_gb->readByte(r.HL);
      DEC(n);
      m_gb->writeByte(r.HL, n);
      break;
    }


    // ### ADD n (A=A+n, F=Z0HC)
    case 0x87: ADD(r.A) break;
    case 0x80: ADD(r.B) break;
    case 0x81: ADD(r.C) break;
    case 0x82: ADD(r.D) break;
    case 0x83: ADD(r.E) break;
    case 0x84: ADD(r.H) break;
    case 0x85: ADD(r.L) break;
    case 0x86:  // ADD A, [HL]
    {
      GBByte n = m_gb->readByte(r.HL);
      ADD(n)
      break;
    }
    case 0xC6:  // ADD A, #v
    {
      GBByte n = m_gb->readByte(r.PC++);
      ADD(n)
      break;
    }


    // ### ADC n (A=A+n+F.c, F=Z0HC)
    case 0x8F: ADC(r.A) break;
    case 0x88: ADC(r.B) break;
    case 0x89: ADC(r.C) break;
    case 0x8A: ADC(r.D) break;
    case 0x8B: ADC(r.E) break;
    case 0x8C: ADC(r.H) break;
    case 0x8D: ADC(r.L) break;
    case 0x8E:  // ADC A, [HL]
    {
      GBByte n = m_gb->readByte(r.HL);
      ADC(n)
      break;
    }
    case 0xCE:  // ADC A, #v
    {
      GBByte n = m_gb->readByte(r.PC++);
      ADC(n)
      break;
    }


    // ### SUB n (A=A-n, F=Z1HC)
    case 0x97: SUB(r.A) break;
    case 0x90: SUB(r.B) break;
    case 0x91: SUB(r.C) break;
    case 0x92: SUB(r.D) break;
    case 0x93: SUB(r.E) break;
    case 0x94: SUB(r.H) break;
    case 0x95: SUB(r.L) break;
    case 0x96:  // SUB A, [HL]
    {
      GBByte n = m_gb->readByte(r.HL);
      SUB(n)
      break;
    }
    case 0xD6:  // SUB A, #v
    {
      GBByte n = m_gb->readByte(r.PC++);
      SUB(n)
      break;
    }


    // ### SBC n (A=A-n-F.c, F=Z1HC)
    case 0x9F: SBC(r.A) break;
    case 0x98: SBC(r.B) break;
    case 0x99: SBC(r.C) break;
    case 0x9A: SBC(r.D) break;
    case 0x9B: SBC(r.E) break;
    case 0x9C: SBC(r.H) break;
    case 0x9D: SBC(r.L) break;
    case 0x9E:  // SBC A, (HL)
    {
      GBByte n = m_gb->readByte(r.HL);
      SBC(n)
      break;
    }
    case 0xDE:  // SBC A, #v
    {
      GBByte n = m_gb->readByte(r.PC++);
      SBC(n)
      break;
    }


    // ### DAA (Decimal adjust register A)
    case 0x27:
    {
      int a = r.A;
      if(r.F & 0x40)
      {
        if(r.F & 0x20)
          a = (a - 6) & 0xFF;
        if(r.F & 0x10)
          a -= 0x60;
      }
      else
      {
        if((r.F & 0x20) || (a & 0x0F) > 9)
            a += 0x06;
        if((r.F & 0x10) || a > 0x9F)
            a += 0x60;
      }
      r.A = static_cast<GBByte>(a);
      r.F = ((r.A == 0) << 7) | (r.F & 0x50) | (((a & 0x100) == 0x100) << 4);
      break;
    }
   

    // ### AND n (A=A&n, F=Z010)
    case 0xA0:  // AND B
      r.A &= r.B;
      r.F = ((r.A == 0) << 7) | (1 << 5);
      break;
    case 0xA1:  // AND C
      r.A &= r.C;
      r.F = ((r.A == 0) << 7) | (1 << 5);
      break;
    case 0xA2:  // AND D
      r.A &= r.D;
      r.F = ((r.A == 0) << 7) | (1 << 5);
      break;
    case 0xA3:  // AND E
      r.A &= r.E;
      r.F = ((r.A == 0) << 7) | (1 << 5);
      break;
    case 0xA4:  // AND H
      r.A &= r.H;
      r.F = ((r.A == 0) << 7) | (1 << 5);
      break;
    case 0xA5:  // AND L
      r.A &= r.L;
      r.F = ((r.A == 0) << 7) | (1 << 5);
      break;
    case 0xA6:  // AND [HL]
      r.A &= m_gb->readByte(r.HL);
      r.F = ((r.A == 0) << 7) | (1 << 5);
      break;
    case 0xE6:  // AND #v
      r.A &=  m_gb->readByte(r.PC++);
      r.F = ((r.A == 0) << 7) | (1 << 5);
      break;
    case 0xA7:  // AND A
      r.F = ((r.A == 0) << 7) | (1 << 5);
      break;



    // ### OR n (A=A|n, F=Z000)
    case 0xB0:  // OR B
      r.A |= r.B;
      r.F = (r.A == 0) << 7;
      break;
    case 0xB1:  // OR C
      r.A |= r.C;
      r.F = (r.A == 0) << 7;
      break;
    case 0xB2:  // OR D
      r.A |= r.D;
      r.F = (r.A == 0) << 7;
      break;
    case 0xB3:  // OR E
      r.A |= r.E;
      r.F = (r.A == 0) << 7;
      break;
    case 0xB4:  // OR H
      r.A |= r.H;
      r.F = (r.A == 0) << 7;
      break;
    case 0xB5:  // OR L
      r.A |= r.L;
      r.F = (r.A == 0) << 7;
      break;
    case 0xB6:  // OR [HL]
      r.A |= m_gb->readByte(r.HL);
      r.F = (r.A == 0) << 7;
      break;
    case 0xF6:  // OR #v
      r.A |= m_gb->readByte(r.PC++);
      r.F = (r.A == 0) << 7;
      break;
    case 0xB7:  // OR A
      r.F = (r.A == 0) << 7;
      break;


    // ### XOR n (A = A^n, F=Z000)
    case 0xAF:  // XOR A
      r.A = 0;
      r.F = 1 << 7;
      break;
    case 0xA8:  // XOR B
      r.A ^= r.B;
      r.F = (r.A == 0) << 7;
      break;
    case 0xA9:  // XOR C
      r.A ^= r.C;
      r.F = (r.A == 0) << 7;
      break;
    case 0xAA:  // XOR D
      r.A ^= r.D;
      r.F = (r.A == 0) << 7;
      break;
    case 0xAB:  // XOR E
      r.A ^= r.E;
      r.F = (r.A == 0) << 7;
      break;
    case 0xAC:  // XOR H
      r.A ^= r.H;
      r.F = (r.A == 0) << 7;
      break;
    case 0xAD:  // XOR L
      r.A ^= r.L;
      r.F = (r.A == 0) << 7;
      break;
    case 0xAE:  // XOR [HL]
      r.A ^= m_gb->readByte(r.HL);
      r.F = (r.A == 0) << 7;
      break;
    case 0xEE:  // XOR #v
      r.A ^= m_gb->readByte(r.PC++);
      r.F = (r.A == 0) << 7;
      break;



    // ### ADD HL, n
    case 0x09:  // ADD HL, BC
      r.F = (r.F & 0x80) | (((r.HL & 0x0FFF) + (r.BC & 0x0FFF) > 0x0FFF) << 5) |
            ((r.HL + r.BC > 0xFFFF) << 4);
      r.HL += r.BC;
      break;
    case 0x19:  // ADD HL, DE
      r.F = (r.F & 0x80) | (((r.HL & 0x0FFF) + (r.DE & 0x0FFF) > 0x0FFF) << 5) |
            ((r.HL + r.DE > 0xFFFF) << 4);
      r.HL += r.DE;
      break;
    case 0x29:  // ADD HL, HL
      r.F = (r.F & 0x80) | (((r.HL & 0x0FFF) > 0x07FF) << 5) |
            ((r.HL > 0x7FFF) << 4);
      r.HL <<= 1;
      break;
    case 0x39:  // ADD HL, SP
      r.F = (r.F & 0x80) | (((r.HL & 0x0FFF) + (r.SP & 0x0FFF) > 0x0FFF) << 5) |
            ((r.HL + r.SP > 0xFFFF) << 4);
      r.HL += r.SP;
      break;



    // ### ADD SP, n
    case 0xE8:
    {
      int v = static_cast<int8_t>(m_gb->readByte(r.PC++));
      if(v >= 0)
        r.F = (((r.SP & 0x0F) + v > 0x0F) << 5) | (((r.SP & 0xFF) + v > 0xFF) << 4);
      else
        r.F = (((r.SP & 0x0F) >= (-v & 0x0F)) << 5) | (((r.SP & 0xFF) >= -v) << 4);
      r.SP += v;
      break;
    }


    // ### CP n  (compares A with N, F = Z1HC)
    case 0xBF:  // CP A
      r.F = ((r.A == r.A) << 7) | 0x40 | (((r.A & 0xF) < (r.A & 0xF)) << 5) |
            ((r.A < r.A) << 4);
      break;
    case 0xB8:  // CP B
      r.F = ((r.A == r.B) << 7) | 0x40 | (((r.A & 0xF) < (r.B & 0xF)) << 5) |
            ((r.A < r.B) << 4);
      break;
    case 0xB9:  // CP C
      r.F = ((r.A == r.C) << 7) | 0x40 | (((r.A & 0xF) < (r.C & 0xF)) << 5) |
            ((r.A < r.C) << 4);
      break;
    case 0xBA:  // CP D
      r.F = ((r.A == r.D) << 7) | 0x40 | (((r.A & 0xF) < (r.D & 0xF)) << 5) |
            ((r.A < r.D) << 4);
      break;
    case 0xBB:  // CP E
      r.F = ((r.A == r.E) << 7) | 0x40 | (((r.A & 0xF) < (r.E & 0xF)) << 5) |
            ((r.A < r.E) << 4);
      break;
    case 0xBC:  // CP H
      r.F = ((r.A == r.H) << 7) | 0x40 | (((r.A & 0xF) < (r.H & 0xF)) << 5) |
            ((r.A < r.H) << 4);
      break;
    case 0xBD:  // CP L
      r.F = ((r.A == r.L) << 7) | 0x40 | (((r.A & 0xF) < (r.L & 0xF)) << 5) |
            ((r.A < r.L) << 4);
      break;
    case 0xBE:  // CP [HL]
    {
      GBByte n = m_gb->readByte(r.HL);
      r.F = ((r.A == n) << 7) | 0x40 | (((r.A & 0xF) < (n & 0xF)) << 5) |
            ((r.A < n) << 4);
      break;
    }
    case 0xFE:  // CP #d
    {
      GBByte n = m_gb->readByte(r.PC++);
      r.F = ((r.A == n) << 7) | 0x40 | (((r.A & 0xF) < (n & 0xF)) << 5) |
            ((r.A < n) << 4);
      break;
    }







    // ### Jumps
    case 0x20:  // JR NZ, #n
      if((r.F & 0x80) == 0)
      {
        dur++;
        addJump(r.PC - 1);
        r.PC += static_cast<int8_t>(m_gb->readByte(r.PC));
      }
      r.PC++;
      break;
    case 0x28:  // JR Z, #n
      if(r.F & 0x80)
      {
        dur++;
        addJump(r.PC - 1);
        r.PC += static_cast<int8_t>(m_gb->readByte(r.PC));
      }
      r.PC++;
      break;
    case 0x30:  // JR NC, #n
      if((r.F & 0x10) == 0)
      {
        dur++;
        addJump(r.PC - 1);
        r.PC += static_cast<int8_t>(m_gb->readByte(r.PC));
      }
      r.PC++;
      break;
    case 0x38:  // JR C, #n
      if(r.F & 0x10)
      {
        dur++;
        addJump(r.PC - 1);
        r.PC += static_cast<int8_t>(m_gb->readByte(r.PC));
      }
      r.PC++;
      break;
    case 0x18:  // JR #n
      addJump(r.PC - 1);
      r.PC += static_cast<int8_t>(m_gb->readByte(r.PC)) + 1;
      break;


    // ### JP
    case 0xC3:  // JP #vv
      addJump(r.PC - 1);
      r.PC = m_gb->readWord(r.PC);
      break;
    case 0xE9:  // JP [HL]
      addJump(r.PC - 1);
      r.PC = r.HL;
      break;
    case 0xC2:  // JP NZ, #vv
      r.PC += 2;
      if((r.F & 0x80) == 0)
      {
        dur++;
        addJump(r.PC - 3);
        r.PC = m_gb->readWord(r.PC - 2);
      }
      break;
    case 0xCA:  // JP Z, #vv
      r.PC += 2;
      if(r.F & 0x80)
      {
        dur++;
        addJump(r.PC - 3);
        r.PC = m_gb->readWord(r.PC - 2);
      }
      break;
    case 0xD2:  // JP NC, #vv
      r.PC += 2;
      if((r.F & 0x10) == 0)
      {
        dur++;
        addJump(r.PC - 3);
        r.PC = m_gb->readWord(r.PC - 2);
      }
      break;
    case 0xDA:  // JP C, #vv
      r.PC += 2;
      if(r.F & 0x10)
      {
        dur++;
        addJump(r.PC - 3);
        r.PC = m_gb->readWord(r.PC - 2);
      }
      break;



    // ### PUSH nn
    case 0xC5:  // PUSH BC
      r.SP -= 2;
      m_gb->writeWord(r.SP, r.BC);
      break;
    case 0xD5:  // PUSH DE
      r.SP -= 2;
      m_gb->writeWord(r.SP, r.DE);
      break;
    case 0xE5:  // PUSH HL
      r.SP -= 2;
      m_gb->writeWord(r.SP, r.HL);
      break;
    case 0xF5:  // PUSH AF
      r.SP -= 2;
      m_gb->writeWord(r.SP, r.AF);
      break;


    // ### POP nn
    case 0xC1:  // POP BC
      r.BC = m_gb->readWord(r.SP);
      r.SP += 2;
      break;
    case 0xD1:  // POP DE
      r.DE = m_gb->readWord(r.SP);
      r.SP += 2;
      break;
    case 0xE1:  // POP HL
      r.HL = m_gb->readWord(r.SP);
      r.SP += 2;
      break;
    case 0xF1:  // POP AF
      r.AF = m_gb->readWord(r.SP);
      r.SP += 2;
      r.F &= 0xF0; // Clear unused flags
      break;





    // ### Calls
    case 0xCD:  // CALL nn
      r.SP -= 2;
      m_gb->writeWord(r.SP, r.PC + 2);
      addJump(r.PC - 1);
      r.PC = m_gb->readWord(r.PC);
      break;
    case 0xC4:  // CALL NZ, nn
      r.PC += 2;
      if((r.F & 0x80) == 0)
      {
        dur += 3;
        r.SP -= 2;
        m_gb->writeWord(r.SP, r.PC);
        addJump(r.PC - 3);
        r.PC = m_gb->readWord(r.PC - 2);
      }
      break;
   case 0xCC:  // CALL Z, nn
      r.PC += 2;
      if(r.F & 0x80)
      {
        dur += 3;
        r.SP -= 2;
        m_gb->writeWord(r.SP, r.PC);
        addJump(r.PC - 3);
        r.PC = m_gb->readWord(r.PC - 2);
      }
      break;
    case 0xD4:  // CALL NC, nn
      r.PC += 2;
      if((r.F & 0x10) == 0)
      {
        dur += 3;
        r.SP -= 2;
        m_gb->writeWord(r.SP, r.PC);
        addJump(r.PC - 3);
        r.PC = m_gb->readWord(r.PC - 2);
      }
      break;
   case 0xDC:  // CALL C, nn
      r.PC += 2;
      if(r.F & 0x10)
      {
        dur += 3;
        r.SP -= 2;
        m_gb->writeWord(r.SP, r.PC);
        addJump(r.PC - 3);
        r.PC = m_gb->readWord(r.PC - 2);
      }
      break;


    // ### RETs
    case 0xC9:  // RET
      addJump(r.PC - 1);
      r.PC = m_gb->readWord(r.SP);
      r.SP += 2;
      break;
    case 0xD9:  // RETI
      m_interruptMasterEnableRegister = 1;
      addJump(r.PC - 1);
      r.PC = m_gb->readWord(r.SP);
      r.SP += 2;
      break;
    case 0xC0:  // RET NZ
      if((r.F & 0x80) == 0)
      {
        dur += 3;
        addJump(r.PC - 1);
        r.PC = m_gb->readWord(r.SP);
        r.SP += 2;
      }
      break;
    case 0xC8:  // RET Z
      if(r.F & 0x80)
      {
        dur += 3;
        addJump(r.PC - 1);
        r.PC = m_gb->readWord(r.SP);
        r.SP += 2;
      }
      break;
    case 0xD0:  // RET NC
      if((r.F & 0x10) == 0)
      {
        dur += 3;
        addJump(r.PC - 1);
        r.PC = m_gb->readWord(r.SP);
        r.SP += 2;
      }
      break;
    case 0xD8:  // RET NZ
      if(r.F & 0x10)
      {
        dur += 3;
        addJump(r.PC - 1);
        r.PC = m_gb->readWord(r.SP);
        r.SP += 2;
      }
      break;


    // ### RST n (push present address onto stack, jump to address 0+n)
    case 0xC7:  // RST 00
      r.SP -= 2;
      m_gb->writeWord(r.SP, r.PC);
      addJump(r.PC - 1);
      r.PC = 0x00;
      break;
    case 0xCF:  // RST 08
      r.SP -= 2;
      m_gb->writeWord(r.SP, r.PC);
      addJump(r.PC - 1);
      r.PC = 0x08;
      break;
    case 0xD7:  // RST 10
      r.SP -= 2;
      m_gb->writeWord(r.SP, r.PC);
      addJump(r.PC - 1);
      r.PC = 0x10;
      break;
    case 0xDF:  // RST 18
      r.SP -= 2;
      m_gb->writeWord(r.SP, r.PC);
      addJump(r.PC - 1);
      r.PC = 0x18;
      break;
    case 0xE7:  // RST 20
      r.SP -= 2;
      m_gb->writeWord(r.SP, r.PC);
      addJump(r.PC - 1);
      r.PC = 0x20;
      break;
    case 0xEF:  // RST 28
      r.SP -= 2;
      m_gb->writeWord(r.SP, r.PC);
      addJump(r.PC - 1);
      r.PC = 0x28;
      break;
    case 0xF7:  // RST 30
      r.SP -= 2;
      m_gb->writeWord(r.SP, r.PC);
      addJump(r.PC - 1);
      r.PC = 0x30;
      break;
    case 0xFF:  // RST 38
      r.SP -= 2;
      m_gb->writeWord(r.SP, r.PC);
      addJump(r.PC - 1);
      r.PC = 0x38;
      break;



    // ### Rotates
    // Note: These are different from their CB-Prefix
    // version since Z-flag is always reset
    case 0x07:  // RLCA
      r.F = (r.A & 0x80) >> 3;
      r.A = r.A << 1 | r.A >> 7;
      break;
    case 0x0F:  // RRCA
      r.F = (r.A & 0x01) << 4;
      r.A = r.A >> 1 | r.A << 7;
      break;
    case 0x17:  // RLA
    {
      bool c = (r.A & 0x80) == 0x80;
      r.A = r.A << 1 | ((r.F & 0x10) >> 4);
      r.F = (c << 4);
      break;
    }
    case 0x1F:  // RRA
    {
      bool c = (r.A & 0x01);
      r.A = r.A >> 1 | ((r.F & 0x10) << 3);
      r.F = (c << 4);
      break;
    }



    // ### Special stuff
    case 0xFB:  // EI (enables interrupts)
      m_interruptMasterEnableRegister = 1;
      break;
    case 0xF3:  // DI (disables interrupts)
      m_interruptMasterEnableRegister = 0;
      break;
    case 0x76:  // HALT
      m_halt = true;
      break;
    case 0x10:  // STOP
      //m_stop = true;
      break;
    case 0x2F:  // CPL (complement/invert A, F=-11-)
      r.A = ~r.A;
      r.F = (r.F & 0x90) | (3 << 5);
      break;
    case 0x37:  // SCF (Set Carry flag)
      r.F = (r.F & 0x80) | (1 << 4);
      break;
    case 0x3F:  // CCF (Complement carry flag)
      r.F = (r.F & 0x80) | (~(r.F & 0x10) & 0x10);
      break;


    // ### CB Prefix
    case 0xCB:
    {
      auto ins2 = m_gb->readByte(r.PC++);
      dur = instructionDurationCB[ins2 & 0x0F];

      // BIT operations (F = Z01-)
      if((ins2 & 0xC0) == 0x40)
      {
        GBByte b = (ins2 & 0x38) >> 3;
        switch(ins2 & 0x07)
        {
          case 0: // BIT B
            r.F = (((r.B & (1 << b)) == 0) << 7) | (1 << 5) | (r.F & 0x10);
            break;
          case 1: // BIT C
            r.F = (((r.C & (1 << b)) == 0) << 7) | (1 << 5) | (r.F & 0x10);
            break;
          case 2: // BIT D
            r.F = (((r.D & (1 << b)) == 0) << 7) | (1 << 5) | (r.F & 0x10);
            break;
          case 3: // BIT E
            r.F = (((r.E & (1 << b)) == 0) << 7) | (1 << 5) | (r.F & 0x10);
            break;
          case 4: // BIT H
            r.F = (((r.H & (1 << b)) == 0) << 7) | (1 << 5) | (r.F & 0x10);
            break;
          case 5: // BIT L
            r.F = (((r.L & (1 << b)) == 0) << 7) | (1 << 5) | (r.F & 0x10);
            break;
          case 6: // BIT [HL]
          {
            dur--;
            GBByte d = m_gb->readByte(r.HL);
            r.F = (((d & (1 << b)) == 0) << 7) | (1 << 5) | (r.F & 0x10);
            break;
          }
          case 7: // BIT A
            r.F = (((r.A & (1 << b)) == 0) << 7) | (1 << 5) | (r.F & 0x10);
            break;
        }
      }
      else if((ins2 & 0xC0) == 0x80) // RES operations
      {
        GBByte b = (ins2 & 0x38) >> 3;
        switch(ins2 & 0x07)
        {
          case 0: // RES B
            r.B &= ~(1 << b);
            break;
          case 1: // RES C
            r.C &= ~(1 << b);
            break;
          case 2: // RES D
            r.D &= ~(1 << b);
            break;
          case 3: // RES E
            r.E &= ~(1 << b);
            break;
          case 4: // RES H
            r.H &= ~(1 << b);
            break;
          case 5: // RES L
            r.L &= ~(1 << b);
            break;
          case 6: // RES [HL]
            m_gb->writeByte(r.HL, m_gb->readByte(r.HL) & ~(1 << b));
            break;
          case 7: // RES A
            r.A &= ~(1 << b);
            break;
        }
      }
      else if((ins2 & 0xC0) == 0xC0) // SET operations
      {
        GBByte b = (ins2 & 0x38) >> 3;
        switch(ins2 & 0x07)
        {
          case 0: // SET B
            r.B |= (1 << b);
            break;
          case 1: // SET C
            r.C |= (1 << b);
            break;
          case 2: // SET D
            r.D |= (1 << b);
            break;
          case 3: // SET E
            r.E |= (1 << b);
            break;
          case 4: // SET H
            r.H |= (1 << b);
            break;
          case 5: // SET L
            r.L |= (1 << b);
            break;
          case 6: // SET [HL]
            m_gb->writeByte(r.HL, m_gb->readByte(r.HL) | (1 << b));
            break;
          case 7: // SET A
            r.A |= (1 << b);
            break;
        }
      }
      else
      {
        switch(ins2)
        {
          // ### RLC n (Rotate n left. Old bit 7 to Carry flag.)
          case 0x00:  // RLC B
          {
            bool c = (r.B & 0x80) == 0x80;
            r.B = r.B << 1 | r.B >> 7;
            r.F = ((r.B == 0) << 7) | (c << 4);
            break;
          }
          case 0x01:  // RLC C
          {
            bool c = (r.C & 0x80) == 0x80;
            r.C = r.C << 1 | r.C >> 7;
            r.F = ((r.C == 0) << 7) | (c << 4);
            break;
          }
          case 0x02:  // RLC D
          {
            bool c = (r.D & 0x80) == 0x80;
            r.D = r.D << 1 | r.D >> 7;
            r.F = ((r.D == 0) << 7) | (c << 4);
            break;
          }
          case 0x03:  // RLC E
          {
            bool c = (r.E & 0x80) == 0x80;
            r.E = r.E << 1 | r.E >> 7;
            r.F = ((r.E == 0) << 7) | (c << 4);
            break;
          }
          case 0x04:  // RLC H
          {
            bool c = (r.H & 0x80) == 0x80;
            r.H = r.H << 1 | r.H >> 7;
            r.F = ((r.H == 0) << 7) | (c << 4);
            break;
          }
          case 0x05:  // RLC L
          {
            bool c = (r.L & 0x80) == 0x80;
            r.L = r.L << 1 | r.L >> 7;
            r.F = ((r.L == 0) << 7) | (c << 4);
            break;
          }
          case 0x06:  // RLC [HL]
          {
            GBByte n = m_gb->readByte(r.HL);
            bool c = (n & 0x80) == 0x80;
            n = n << 1 | n >> 7;
            r.F = ((n == 0) << 7) | (c << 4);
            m_gb->writeByte(r.HL, n);
            break;
          }
          case 0x07:  // RLC A
          {
            bool c = (r.A & 0x80) == 0x80;
            r.A = r.A << 1 | r.A >> 7;
            r.F = ((r.A == 0) << 7) | (c << 4);
            break;
          }


          // ### RRC n (Rotate n right. Old bit 0 to Carry flag.)
          case 0x08:  // RRC B
          {
            bool c = (r.B & 0x01);
            r.B = r.B >> 1 | r.B << 7;
            r.F = ((r.B == 0) << 7) | (c << 4);
            break;
          }
          case 0x09:  // RRC C
          {
            bool c = (r.C & 0x01);
            r.C = r.C >> 1 | r.C << 7;
            r.F = ((r.C == 0) << 7) | (c << 4);
            break;
          }
          case 0x0A:  // RRC D
          {
            bool c = (r.D & 0x01);
            r.D = r.D >> 1 | r.D << 7;
            r.F = ((r.D == 0) << 7) | (c << 4);
            break;
          }
          case 0x0B:  // RRC E
          {
            bool c = (r.E & 0x01);
            r.E = r.E >> 1 | r.E << 7;
            r.F = ((r.E == 0) << 7) | (c << 4);
            break;
          }
          case 0x0C:  // RRC H
          {
            bool c = (r.H & 0x01);
            r.H = r.H >> 1 | r.H << 7;
            r.F = ((r.H == 0) << 7) | (c << 4);
            break;
          }
          case 0x0D:  // RRC L
          {
            bool c = (r.L & 0x01);
            r.L = r.L >> 1 | r.L << 7;
            r.F = ((r.L == 0) << 7) | (c << 4);
            break;
          }
          case 0x0E:  // RRC [HL]
          {
            GBByte n = m_gb->readByte(r.HL);
            bool c = (n & 0x01);
            n = n >> 1 | n << 7;
            r.F = ((n == 0) << 7) | (c << 4);
            m_gb->writeByte(r.HL, n);
            break;
          }
          case 0x0F:  // RRC A
          {
            bool c = (r.A & 0x01);
            r.A = r.A >> 1 | r.A << 7;
            r.F = ((r.A == 0) << 7) | (c << 4);
            break;
          }

          // ### RL n (Rotate n left through Carry flag.)
          case 0x10:  // RL B
          {
            bool c = (r.B & 0x80) == 0x80;
            r.B = r.B << 1 | ((r.F & 0x10) >> 4);
            r.F = ((r.B == 0) << 7) | (c << 4);
            break;
          }
          case 0x11:  // RL C
          {
            bool c = (r.C & 0x80) == 0x80;
            r.C = r.C << 1 | ((r.F & 0x10) >> 4);
            r.F = ((r.C == 0) << 7) | (c << 4);
            break;
          }
          case 0x12:  // RL D
          {
            bool c = (r.D & 0x80) == 0x80;
            r.D = r.D << 1 | ((r.F & 0x10) >> 4);
            r.F = ((r.D == 0) << 7) | (c << 4);
            break;
          }
          case 0x13:  // RL E
          {
            bool c = (r.E & 0x80) == 0x80;
            r.E = r.E << 1 | ((r.F & 0x10) >> 4);
            r.F = ((r.E == 0) << 7) | (c << 4);
            break;
          }
          case 0x14:  // RL H
          {
            bool c = (r.H & 0x80) == 0x80;
            r.H = r.H << 1 | ((r.F & 0x10) >> 4);
            r.F = ((r.H == 0) << 7) | (c << 4);
            break;
          }
          case 0x15:  // RL L
          {
            bool c = (r.L & 0x80) == 0x80;
            r.L = r.L << 1 | ((r.F & 0x10) >> 4);
            r.F = ((r.L == 0) << 7) | (c << 4);
            break;
          }
          case 0x16:  // RL [HL]
          {
            GBByte n = m_gb->readByte(r.HL);
            bool c = (n & 0x80) == 0x80;
            n = n << 1 | ((r.F & 0x10) >> 4);
            r.F = ((n == 0) << 7) | (c << 4);
            m_gb->writeByte(r.HL, n);
            break;
          }
          case 0x17:  // RL A
          {
            bool c = (r.A & 0x80) == 0x80;
            r.A = r.A << 1 | ((r.F & 0x10) >> 4);
            r.F = ((r.A == 0) << 7) | (c << 4);
            break;
          }


          // ### RR n (Rotate n right through Carry flag.)
          case 0x18:  // RR B
          {
            bool c = (r.B & 0x01);
            r.B = r.B >> 1 | ((r.F & 0x10) << 3);
            r.F = ((r.B == 0) << 7) | (c << 4);
            break;
          }
          case 0x19:  // RR C
          {
            bool c = (r.C & 0x01);
            r.C = r.C >> 1 | ((r.F & 0x10) << 3);
            r.F = ((r.C == 0) << 7) | (c << 4);
            break;
          }
          case 0x1A:  // RR D
          {
            bool c = (r.D & 0x01);
            r.D = r.D >> 1 | ((r.F & 0x10) << 3);
            r.F = ((r.D == 0) << 7) | (c << 4);
            break;
          }
          case 0x1B:  // RR E
          {
            bool c = (r.E & 0x01);
            r.E = r.E >> 1 | ((r.F & 0x10) << 3);
            r.F = ((r.E == 0) << 7) | (c << 4);
            break;
          }
          case 0x1C:  // RR H
          {
            bool c = (r.H & 0x01);
            r.H = r.H >> 1 | ((r.F & 0x10) << 3);
            r.F = ((r.H == 0) << 7) | (c << 4);
            break;
          }
          case 0x1D:  // RR L
          {
            bool c = (r.L & 0x01);
            r.L = r.L >> 1 | ((r.F & 0x10) << 3);
            r.F = ((r.L == 0) << 7) | (c << 4);
            break;
          }
          case 0x1E:  // RR [HL]
          {
            GBByte n = m_gb->readByte(r.HL);
            bool c = (n & 0x01);
            n = n >> 1 | ((r.F & 0x10) << 3);
            r.F = ((n == 0) << 7) | (c << 4);
            m_gb->writeByte(r.HL, n);
            break;
          }
          case 0x1F:  // RR A
          {
            bool c = (r.A & 0x01);
            r.A = r.A >> 1 | ((r.F & 0x10) << 3);
            r.F = ((r.A == 0) << 7) | (c << 4);
            break;
          }


          // ### SLA n (Shift n left into Carry. LSB of n set to 0.)
          case 0x20:  // SLA B
          {
            bool c = (r.B & 0x80) == 0x80;
            r.B = r.B << 1;
            r.F = ((r.B == 0) << 7) | (c << 4);
            break;
          }
          case 0x21:  // SLA C
          {
            bool c = (r.C & 0x80) == 0x80;
            r.C = r.C << 1;
            r.F = ((r.C == 0) << 7) | (c << 4);
            break;
          }
          case 0x22:  // SLA D
          {
            bool c = (r.D & 0x80) == 0x80;
            r.D = r.D << 1;
            r.F = ((r.D == 0) << 7) | (c << 4);
            break;
          }
          case 0x23:  // SLA E
          {
            bool c = (r.E & 0x80) == 0x80;
            r.E = r.E << 1;
            r.F = ((r.E == 0) << 7) | (c << 4);
            break;
          }
          case 0x24:  // SLA H
          {
            bool c = (r.H & 0x80) == 0x80;
            r.H = r.H << 1;
            r.F = ((r.H == 0) << 7) | (c << 4);
            break;
          }
          case 0x25:  // SLA L
          {
            bool c = (r.L & 0x80) == 0x80;
            r.L = r.L << 1;
            r.F = ((r.L == 0) << 7) | (c << 4);
            break;
          }
          case 0x26:  // SLA [HL]
          {
            GBByte n = m_gb->readByte(r.HL);
            bool c = (n & 0x80) == 0x80;
            n = n << 1;
            r.F = ((n == 0) << 7) | (c << 4);
            m_gb->writeByte(r.HL, n);
            break;
          }
          case 0x27:  // SLA A
          {
            bool c = (r.A & 0x80) == 0x80;
            r.A = r.A << 1;
            r.F = ((r.A == 0) << 7) | (c << 4);
            break;
          }


          // ### SRA n (Shift n right into Carry. MSB doesn't change.)
          case 0x28:  // SRA B
          {
            bool c = (r.B & 0x01);
            r.B = r.B >> 1 | (r.B & 0x80);
            r.F = ((r.B == 0) << 7) | (c << 4);
            break;
          }
          case 0x29:  // SRA C
          {
            bool c = (r.C & 0x01);
            r.C = r.C >> 1 | (r.C & 0x80);
            r.F = ((r.C == 0) << 7) | (c << 4);
            break;
          }
          case 0x2A:  // SRA D
          {
            bool c = (r.D & 0x01);
            r.D = r.D >> 1 | (r.D & 0x80);
            r.F = ((r.D == 0) << 7) | (c << 4);
            break;
          }
          case 0x2B:  // SRA E
          {
            bool c = (r.E & 0x01);
            r.E = r.E >> 1 | (r.E & 0x80);
            r.F = ((r.E == 0) << 7) | (c << 4);
            break;
          }
          case 0x2C:  // SRA H
          {
            bool c = (r.H & 0x01);
            r.H = r.H >> 1 | (r.H & 0x80);
            r.F = ((r.H == 0) << 7) | (c << 4);
            break;
          }
          case 0x2D:  // SRA L
          {
            bool c = (r.L & 0x01);
            r.L = r.L >> 1 | (r.L & 0x80);
            r.F = ((r.L == 0) << 7) | (c << 4);
            break;
          }
          case 0x2E:  // SRA [HL]
          {
            GBByte n = m_gb->readByte(r.HL);
            bool c = (n & 0x01);
            n = n >> 1 | (n & 0x80);
            r.F = ((n == 0) << 7) | (c << 4);
            m_gb->writeByte(r.HL, n);
            break;
          }
          case 0x2F:  // SRA A
          {
            bool c = (r.A & 0x01);
            r.A = r.A >> 1 | (r.A & 0x80);
            r.F = ((r.A == 0) << 7) | (c << 4);
            break;
          }


          // ### SWAP n (swap nibbles, F=Z000)
          case 0x30:  // swap B
            r.B = (r.B << 4) | (r.B >> 4);
            r.F = (r.B == 0) << 7;
            break;
          case 0x31:  // swap C
            r.C = (r.C << 4) | (r.C >> 4);
            r.F = (r.C == 0) << 7;
            break;
          case 0x32:  // swap D
            r.D = (r.D << 4) | (r.D >> 4);
            r.F = (r.D == 0) << 7;
            break;
          case 0x33:  // swap E
            r.E = (r.E << 4) | (r.E >> 4);
            r.F = (r.E == 0) << 7;
            break;
          case 0x34:  // swap H
            r.H = (r.H << 4) | (r.H >> 4);
            r.F = (r.H == 0) << 7;
            break;
          case 0x35:  // swap L
            r.L = (r.L << 4) | (r.L >> 4);
            r.F = (r.L == 0) << 7;
            break;
          case 0x36:  // swap [HL]
          {
            GBByte d = m_gb->readByte(r.HL);
            d = (d << 4) | (d >> 4);
            r.F = (d == 0) << 7;
            m_gb->writeByte(r.HL, d);
            break;
          }
          case 0x37:  // swap A
            r.A = (r.A << 4) | (r.A >> 4);
            r.F = (r.A == 0) << 7;
            break;


          // ### SRL n (Shift n right into Carry. MSB set to 0.)
          case 0x38:  // SRL B
          {
            bool c = (r.B & 0x01);
            r.B = r.B >> 1;
            r.F = ((r.B == 0) << 7) | (c << 4);
            break;
          }
          case 0x39:  // SRL C
          {
            bool c = (r.C & 0x01);
            r.C = r.C >> 1;
            r.F = ((r.C == 0) << 7) | (c << 4);
            break;
          }
          case 0x3A:  // SRL D
          {
            bool c = (r.D & 0x01);
            r.D = r.D >> 1;
            r.F = ((r.D == 0) << 7) | (c << 4);
            break;
          }
          case 0x3B:  // SRL E
          {
            bool c = (r.E & 0x01);
            r.E = r.E >> 1;
            r.F = ((r.E == 0) << 7) | (c << 4);
            break;
          }
          case 0x3C:  // SRL H
          {
            bool c = (r.H & 0x01);
            r.H = r.H >> 1;
            r.F = ((r.H == 0) << 7) | (c << 4);
            break;
          }
          case 0x3D:  // SRL L
          {
            bool c = (r.L & 0x01);
            r.L = r.L >> 1;
            r.F = ((r.L == 0) << 7) | (c << 4);
            break;
          }
          case 0x3E:  // SRL [HL]
          {
            GBByte n = m_gb->readByte(r.HL);
            bool c = (n & 0x01);
            n = n >> 1;
            r.F = ((n == 0) << 7) | (c << 4);
            m_gb->writeByte(r.HL, n);
            break;
          }
          case 0x3F:  // SRL A
          {
            bool c = (r.A & 0x01);
            r.A = r.A >> 1;
            r.F = ((r.A == 0) << 7) | (c << 4);
            break;
          }


          default:
            cout << "Could not execute: CB " << hex << +ins2 << dec << endl;
            throw runtime_error("Could not execute instruction!");
        }
      }
      break;
    }

    // ### invalid opcodes!
    case 0xD3:
    case 0xE3:
    case 0xE4:
    case 0xF4:
    case 0xDB:
    case 0xEB:
    case 0xEC:
    case 0xFC:
    case 0xDD:
    case 0xED:
    case 0xFD:
      cout << FG_RED << "!!! INVALID OPCODE! (" << hex << +ins << dec
           << ")" << C_RESET_FG << endl;
      break;

    default:
      cout << FG_RED << "Could not execute: " << hex << +ins
           << dec << C_RESET_FG << endl;
      throw runtime_error("Could not execute instruction!");
  }

  return MachineCycle(dur);
}



std::ostream& operator<<(std::ostream& out, const Processor::Register& reg)
{
  out << setfill('0') << hex;
  out << "A: " << setw(2) << +reg.A
      << ", F(ZAHC): " << ((reg.F & 0x80) == 0x80) << ((reg.F & 0x40) == 0x40)
      << ((reg.F & 0x20) == 0x20) << ((reg.F & 0x10) == 0x10);
  out << ", BC: " << setw(4) << +reg.BC
      << ", DE: " << setw(4) << +reg.DE
      << ", HL: " << setw(4) << +reg.HL
      << ", SP: " << setw(4) << +reg.SP
      << ", PC: " << setw(4) << +reg.PC;
  return out;
}
