#pragma once

#include "Types.hpp"

#include "../libs/InOutAlgo.hpp"

#include <cstdint>
#include <memory>
#include <iostream>
#include <iomanip>
#include <vector>


class GameBoy;

enum INTERRUPT_BITS
{
  INTERRUPT_VBLANK    = 0,
  INTERRUPT_LCDC_STAT = 1,
  INTERRUPT_TIMER     = 2,
  INTERRUPT_SERIAL    = 3,
  INTERRUPT_JOYPAD    = 4
};

class Processor
{
public:
  Processor(GameBoy* gb);
  ~Processor() {}

  GBByte* getInternalRom() const
  {
    return m_startUpRom.get();
  }

  void checkInterrupts();
  MachineCycle cycle();

  // Debug stuff
  GBPtr getPC() const
  {
    return m_register.PC;
  }
  GBPtr getSP() const
  {
    return m_register.SP;
  }
  void addDebugCommands();

  void printLastJumps() const
  {
    using namespace std;
    using namespace ioa;
    if(m_lastJumpSources.size() == 0)
      return;

    for(size_t i = (m_lastJumpSources.size() - 20) % m_lastJumpSources.size();
        i < m_lastJumpSources.size(); i++)
    {
      cout << setfill('0') << setw(4) << hex << FG_BLUE << m_lastJumpSources[i] 
           << C_RESET_FG << " -> ";
    }
    cout << FG_BLUE << m_register.PC << C_RESET_FG << endl;
  }

  void addJump(GBPtr ptr) 
  {
    if(m_lastJumpSources.size() == 0 ||
       m_lastJumpSources[m_lastJumpSources.size() - 1] != ptr)
      m_lastJumpSources.push_back(ptr);
  }

  GBByte getNextIns() const;
  void printRegisters() const
  {
    std::cout << ioa::FG_BLUE << m_register << ioa::C_RESET_FG << std::endl;
  }

private:
  struct Register
  {
    union   // accumulator and flags
    {
      GBWord AF;
      struct { GBByte F; GBByte A; };
    };
    union   // general purpose register 1
    {
      GBWord BC;
      struct { GBByte C; GBByte B; };
    };
    union   // general purpose register 2
    {
      GBWord DE;
      struct { GBByte E; GBByte D; };
    };
    union   // general purpose register 3
    {
      GBWord HL;
      struct { GBByte L; GBByte H; };
    };
    GBWord SP;  // stack pointer
    GBWord PC;  // program counter
  };

  enum FlagOffset
  {
    FO_CarryFlag = 4,
    FO_HalfCarryFlag = 5,
    FO_AddSubFlag = 6,
    FO_ZeroFlag = 7
  };


  GameBoy* m_gb;
  Register m_register;
  GBByte m_interruptMasterEnableRegister;
  bool m_halt, m_stop;
  std::unique_ptr<GBByte[]> m_startUpRom;
  std::vector<GBPtr> m_lastJumpSources;

  friend std::ostream& operator<<(std::ostream& out,
                                  const Processor::Register& reg);
  friend inline std::ostream& operator<<(std::ostream& out,
                                         const Processor& cpu);
};

std::ostream& operator<<(std::ostream& out, const Processor::Register& reg);
inline std::ostream& operator<<(std::ostream& out, const Processor& cpu)
{
  return out << cpu.m_register;
}

