/**
* InOutAlgo
* Small library for some more or less useful functions and helper objects.
* \author Lukas Kalbertodt <lukas.kalbertodt@gmail.com>
* \version 0.1.1
*
*/

// idempotency guard
#ifndef in_out_algo_hpp
#define in_out_algo_hpp

// prevent stupid microsoft macros (and reset define state at end of this header)
#ifndef NOMINMAX
  #define NOMINMAX
  #define _IOA_RESET_MINMAX
#endif


// STL header
#include <iostream>
#include <string>
#include <stdint.h>
#include <typeinfo>
#include <limits>

// OS dependend header (for colour)
#ifdef _WIN32
  #include <windows.h>
#else  // _WIN32
  #include <sys/stat.h>
  #include <sys/types.h>
#endif  // _WIN32

namespace ioa
{
	// ===========================================================================
	// ####################################################						||
	// ### function and class declarations ################						||
	// ####################################################						||
	// ===========================================================================
	// stores rgba colors, convertion functions for terminal color 
	class Colour;

	// pre definied foreground colours
	enum ConsoleFG //: uint32_t
	{
		FG_RED = 0xFF0000FF,
		FG_GREEN = 0x00FF00FF,
		FG_BLUE = 0x0000FFFF,
		FG_YELLOW = 0xFFFF00FF,
		FG_TURQUOISE = 0x00FFFFFF,
		FG_PURPLE = 0xFF00FFFF,
		FG_WHITE = 0xFFFFFFFF,
		FG_BLACK = 0x000000FF
	};

	// pre definied background colours
	enum ConsoleBG //: uint32_t
	{
		BG_RED = 0xFF0000FF,
		BG_GREEN = 0x00FF00FF,
		BG_BLUE = 0x0000FFFF,
		BG_YELLOW = 0xFFFF00FF,
		BG_TURQUOISE = 0x00FFFFFF,
		BG_PURPLE = 0xFF00FFFF,
		BG_WHITE = 0xFFFFFFFF,
		BG_BLACK = 0x000000FF
	};

	// special values to reset colours
	enum ConsoleSpecial
	{
		C_RESET,
		C_RESET_FG,
		C_RESET_BG
	};

	// manipulate color output with cout
	inline std::ostream& operator<< (std::ostream&, const ConsoleFG);
	inline std::ostream& operator<< (std::ostream&, const ConsoleBG);
	inline std::ostream& operator<< (std::ostream&, const ConsoleSpecial);

	// simple type traits for readable type names
	template <typename T> struct TypeNames;

	// type trait to correctly resolve every type (with C++11)
	template <typename T> struct TypeHelper;

#ifdef IOA_CPP11
	// type traits for resolving array types
	template <typename T, char... Cs> struct ArrayBase;

	// type traits for resolving ptr + ref types
	template <typename T, char... Cs> struct TypeBase;

	// function to explain a type in englisch language
	template <typename T> void ExplainType();
	template <typename T> void ExplainType(T&);

	// helper function for ExplainType to get "a" / "an"
	inline const char* GetArticle(char first);
#endif

	// type traits for simply resolve easy array types (c++03)
	template <typename T> struct ArrayTraits;

	// function to read input from user (simple or array)
	template <typename T> T GetInput();
	template <typename T> void GetInput(T& array, const std::size_t len);
	template <typename T> void GetInput(T& array);





	// ===========================================================================
	// ####################################################						||
	// ### color enum and class ###########################						||
	// ####################################################						||
	// ===========================================================================

	// class to store rgba colors and konvert it to console colors
	class Colour
	{
	public:
		Colour()
			: r(0), g(0), b(0), a(255)
		{}

		Colour(uint8_t r, uint8_t g, uint8_t b, uint8_t a = 255)
			: r(r), g(g), b(b), a(a)
		{}

		Colour(ConsoleFG c)
			: r((c & 0xFF000000) >> 24),
			g((c & 0x00FF0000) >> 16),
			b((c & 0x0000FF00) >> 8),
			a(c & 0x000000FF)
		{}

		Colour(ConsoleBG c)
			: r((c & 0xFF000000) >> 24),
			g((c & 0x00FF0000) >> 16),
			b((c & 0x0000FF00) >> 8),
			a(c & 0x000000FF)
		{}



		uint8_t GetWinForeground(const Colour& other = Colour()) const
		{
			uint8_t out = ~other.a & 0x80;	// invert bg intensity
			out |= (other.r & 0x80) >> 1;
			out |= (other.g & 0x80) >> 2;
			out |= (other.b & 0x80) >> 3;
			out |= (a & 0x80) >> 4;
			out |= (r & 0x80) >> 5;
			out |= (g & 0x80) >> 6;
			out |= (b & 0x80) >> 7;
			return out;
		}

		uint8_t GetWinBackground(const Colour& other = Colour(255, 255, 255)) const
		{
			return other.GetWinForeground(*this);
		}

		void GetLinuxCompleteColor(char* out) const
		{
			out[0] = a >= 128 ? '1' : '2';
			out[1] = ';';
			out[2] = '3';
			out[3] = GetLinuxSingleColor();
		}

		char GetLinuxSingleColor() const
		{
			bool br = r >= 128;
			bool bg = g >= 128;
			bool bb = b >= 128;
			if(br && bg && bb) 			return '7';
			if(!br && !bg && !bb) 	return '0';
			if(br && !bg && !bb)	return '1';
			if(!br && bg && !bb)	return '2';
			if(br && bg && !bb)	return '3';
			if(!br && !bg && bb)	return '4';
			if(br && !bg && bb)	return '5';
			if(!br && bg && bb)	return '6';
			return '0';
		}

		ConsoleFG GetAsForeground()
		{
			return static_cast<ConsoleFG>(
				(r >= 128)*0xFF000000
				| (g >= 128)*0x00FF0000
				| (b >= 128)*0x0000FF00
				| (a >= 128)*0x000000FF);
		}

		ConsoleBG GetAsBackground()
		{
			return static_cast<ConsoleBG>(GetAsForeground());
		}

	private:
		uint8_t r, g, b, a;
	};

	// manipulate foreground colour with cout
	inline std::ostream& operator<< (std::ostream& out, const ConsoleFG cc)
	{
		Colour c(cc);
#ifdef _WIN32
		CONSOLE_SCREEN_BUFFER_INFO csbi;
		GetConsoleScreenBufferInfo(GetStdHandle(STD_OUTPUT_HANDLE), &csbi );
		SetConsoleTextAttribute(GetStdHandle(STD_OUTPUT_HANDLE), 
			(csbi.wAttributes & 0xFFF0) | c.GetWinForeground());
#else
		char code[] = "\033[0000m";
		c.GetLinuxCompleteColor(code + 2);
		out << code;
#endif
		return out;
	}

	// manipulate background colour with cout
	inline std::ostream& operator<< (std::ostream& out, const ConsoleBG cc)
	{
		Colour c(cc);
#ifdef _WIN32
		CONSOLE_SCREEN_BUFFER_INFO csbi;
		GetConsoleScreenBufferInfo(GetStdHandle(STD_OUTPUT_HANDLE), &csbi );
		SetConsoleTextAttribute(GetStdHandle(STD_OUTPUT_HANDLE), 
			(csbi.wAttributes & 0xFF0F) | c.GetWinBackground());
#else
		char code[] = "\033[40m";
		code[3] = c.GetLinuxSingleColor();
		out << code;
#endif
		return out;
	}

	// manipulate colour with special values
	inline std::ostream& operator<< (std::ostream& out, const ConsoleSpecial s)
	{
#ifdef _WIN32
		if(s == C_RESET)
		{
			out << FG_WHITE;
			out << BG_BLACK;
		}
		else if(s == C_RESET_FG)
			out << FG_WHITE;
		else if(s == C_RESET_BG)
			out << BG_BLACK;
#else
		if(s == C_RESET)
			out << "\033[0m";
		else if(s == C_RESET_FG)
			out << "\033[22;39m";
		else if(s == C_RESET_BG)
			out << "\033[49m";
#endif
		return out;
	}





	// ===========================================================================
	// ####################################################						||
	// ### type traits ####################################						||
	// ####################################################						||
	// ===========================================================================

	// --- simple type names
	template <typename T>
	struct TypeNames {
		static const char* name() { return typeid(T).name(); }
	};
	// specialization for standard types
	template<> struct TypeNames<int> {
		static const char* name() { return "int"; }
	};
	template<> struct TypeNames<float> {
		static const char* name() { return "float"; }
	};
	template<> struct TypeNames<double> {
		static const char* name() { return "double"; }
	};
	template<> struct TypeNames<short> {
		static const char* name() { return "short"; }
	};
	template<> struct TypeNames<char> {
		static const char* name() { return "char"; }
	};
	template<> struct TypeNames<bool> {
		static const char* name() { return "bool"; }
	};
	template<> struct TypeNames<std::string> {
		static const char* name() { return "string"; }
	};
	template<> struct TypeNames<void> {
		static const char* name() { return "void"; }
	};

#ifdef IOA_CPP11
	// --- class for resolving ptr, ref + const types
	template <typename T, char... Cs>
	struct TypeBase
	{
		const char* name() { return TypeNames<T>::name(); }
		typedef T core_type;
		const char suffix[sizeof...(Cs) + 1] = { Cs... };
	};

	// partial specialization
	template <typename T, char... Cs>
	struct TypeBase<T*, Cs...> : public ArrayBase<T, '*', Cs...> {
		static const bool is_pointer = true;
	};
	template <typename T, char... Cs>
	struct TypeBase<T&, Cs...> : public ArrayBase<T, '&', Cs...> {
	};
	template <typename T, char... Cs>
	struct TypeBase<T&&, Cs...> : public ArrayBase<T, '&', '&', Cs...> {
	};
	template <typename T, char... Cs>
	struct TypeBase<const T, Cs...> :public TypeBase<T, 'c', Cs...> {
	};


	// --- class for resolving array types
	template <typename T, char... Cs> 
	struct ArrayBase : public TypeBase<T, Cs...>{
	};

	// partial specialization
	template <typename T, char... Cs>
	struct ArrayBase<T[], Cs...> : public TypeBase<T, '[', ']', Cs...>{
	};
	template <typename T, std::size_t N, char... Cs>
	struct ArrayBase<T[N], Cs...> : public TypeBase<T, '[', N + '0', ']', Cs...>{
	};


	// --- helper struct for wrapping base structs
	// is defined in another way if c++11 is not available
	// Be carefull! Do not include this file twice with different
	// C++11 settings!
	template <typename T>
	struct TypeHelper : public ArrayBase<T> {	};

	// --- function to explain a type in english language
	template <typename T>
	void ExplainType(T&)
	{
		ExplainType<T>();
	}

	template <typename T>
	void ExplainType()
	{
		TypeHelper<T> th;
		std::cout << "T is ";

		// parse suffix
		bool articleSet = false;
		for(int i = sizeof(th.suffix) - 1; i >= 0; i--)
		{
			switch(th.suffix[i])
			{
			case ']':
				{
					if(th.suffix[i-1] == '[')
					{
						std::cout << "an array of ";
						i--;
					}
					else
					{
						std::cout << "an array with size '" 
							<< th.suffix[i-1]
						<< "' of ";
						i -= 2; 
					}
					break;
				}
			case '*':
				{
					if(!articleSet)
						std::cout << "a ";
					std::cout << "pointer to ";
					articleSet = false;
					break;
				}
			case '&':
				{
					if(!articleSet)
						std::cout << "a ";
					std::cout << "reference to ";
					articleSet = false;
					break;
				}
			case 'c':
				{
					std::cout << "a const ";
					articleSet = true;
					break;
				}
			}
		}


		// core type
		const char* core = th.name();
		if(!articleSet)
			std::cout << GetArticle(*core) << " ";
		std::cout << core << std::endl;
	}

	// helper function for ExplainType
	inline const char* GetArticle(char first)
	{
		if(first == 'a' || first == 'e' || first == 'i' || first == 'o' || first == 'u')
			return "an";
		return "a";
	}
#else
	// --- if no c++11 available, TypeHelper should just use TypeNames
	// is defined in another way if c++11 is not available
	// Be carefull! Do not include this file twice with different
	// C++11 settings!
	template <typename T>
	struct TypeHelper : public TypeNames<T> {	
		const char* suffix;

		TypeHelper() : suffix("") {}
	};

#endif


	// ===========================================================================
	// ####################################################						||
	// ### functions for reading user input ###############						||
	// ####################################################						||
	// ===========================================================================

	// --- simple array traits
	template <typename T>
	struct ArrayTraits {
		typedef T core_type;
		static const int array_size = 0;
	};
	template <typename T, std::size_t N>
	struct ArrayTraits<T[N]> {
		typedef T core_type;
		static const int array_size = N;
	};
	template <typename Tx, typename Ty>
	struct is_same {
		static const bool value = false;
	};
	template <typename T>
	struct is_same<T,T> {
		static const bool value = true;
	};

	// -- helper function 
	template <typename T>
	void _ReadFromCin(T& v)
	{
		std::cin >> v;
	}
	template <>
	inline void _ReadFromCin(std::string& v)
	{
		std::getline(std::cin, v);
	}

	inline std::streamsize _CleanLine(bool alwaysClean = false)
	{
		std::streamsize cread = 0;
		if(alwaysClean 
			|| (std::cin.rdbuf()->sungetc() != std::char_traits<char>::eof()
			&& std::cin.get() != std::cin.widen('\n') 
			) 
			)
		{
			std::cin.ignore(std::numeric_limits<std::streamsize>::max(), std::cin.widen('\n'));
			cread = std::cin.gcount();
		}
		return cread;
	}

	inline void _ClearCin()
	{
		// clear stream states
		std::cin.clear();
		_CleanLine();
	}

	// --- simple data types
	template <typename T>
	T GetInput()
	{
		T obj;

		while(true)
		{
			_ClearCin();
			// read formatted
			std::cout << "Bitte ein " 
				<< FG_GREEN
				<< TypeHelper<T>().name() 
				<< TypeHelper<T>().suffix
				<< C_RESET_FG
				<< " eingeben:   " << std::flush;

			_ReadFromCin(obj);

			// if everything is ok
			if(std::cin)
				break;

			// error occured -> reset and try again
			std::cout << FG_RED << "Fehler bei der Eingabe!" << C_RESET << std::endl;
		}
		_ClearCin();
		return obj;
	}

	// --- array types
	template <typename T>
	void GetInput(T& array, const std::size_t len)
	{
		const char* tname = TypeNames<typename ArrayTraits<T>::core_type>::name();
		std::cout << "Bitte " << FG_BLUE << len << C_RESET_FG << " "
			<< FG_GREEN << tname << C_RESET_FG << " eingeben." << std::endl;
		for(std::size_t i = 0; i < len; i++)
		{
			while(true)
			{
				_ClearCin();
				std::cout << FG_GREEN << tname << C_RESET_FG
					<< " (" << FG_TURQUOISE << i+1 << C_RESET_FG
					<< "/" << FG_BLUE << len << C_RESET_FG << ") eingeben:   ";
				_ReadFromCin(array[i]);

				// if everything is ok
				if(std::cin)
					break;

				// error occured -> reset and try again
				std::cout << FG_RED << "Fehler bei der Eingabe!" << C_RESET << std::endl;

				// clear stream states
				std::cin.clear();

				// clear stream (manually high number to prevent from
				// including a new header only for that)
				std::cin.ignore(0x0FFFFFFF, '\n');
			}
			_ClearCin();
		}
	}

	// array types
	template <typename T>
	void GetInput(T& array)
	{
#ifdef IOA_CPP11
		static_assert(ArrayTraits<T>::array_size > 0, 
			"This Overload is made for arrays, not primitive data types!");
#endif
		GetInput(array, ArrayTraits<T>::array_size);
	}


};

#ifdef _IOA_RESET_MINMAX
  #undef NOMINMAX
  #undef _IOA_RESET_MINMAX
#endif  // _IOA_RESET_MINMAX

#endif // in_out_algo_hpp